#!/usr/bin/env python3
#
# Tests Myokit's SBML support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats
import myokit.formats.sbml
from myokit.formats.sbml import SBMLParser, SBMLParsingError

from myokit.tests import WarningCollector


class TestSBMLParser(unittest.TestCase):
    """
    Unit tests for the SBMLParser class.
    """

    @classmethod
    def setUpClass(cls):
        cls.p = SBMLParser()

    def assertBad(self, xml, message, lvl=3, v=2):
        """
        Inserts the given ``xml`` into a <model> element, parses it, and checks
        that this raises an exception matching ``message``.
        """
        self.assertRaisesRegex(
            SBMLParsingError, message, self.parse, xml, lvl, v)

    def parse(self, xml, lvl=3, v=2):
        """
        Inserts the given ``xml`` into an <sbml> element, parses it, and
        returns the result.
        """
        return self.p.parse_string(self.wrap(xml, lvl, v))

    def wrap(self, xml_content, level=3, version=2):
        """
        Wraps ``xml_content`` into an SBML document of the specified ``level``
        and ``version``.
        """
        lv = 'level' + str(level) + '/version' + str(version)
        return (
            '<sbml xmlns="http://www.sbml.org/sbml/' + lv + '/core"'
            ' level="' + str(level) + '"'
            ' version="' + str(version) + '">'
            + xml_content +
            '</sbml>'
        )

    def test_duplicate_sids(self):
        # Checks that an error is thrown when an SId is used twice.
        # Nearly every object can have an SId, but Myokit only checks:
        #  compartments, species, species references, reactions, parameters
        # Units do not have SIds, but their own UnitSIds.

        # Coinciding compartment and parameter ids
        xml = (
            '<model id="test" '
            'timeUnits="second">'
            '<listOfCompartments>'
            '<compartment id="someId"/>'
            '</listOfCompartments>'
            '<listOfParameters>'
            '<parameter id="someId"/>'
            '</listOfParameters>'
            '</model>')
        self.assertBad(xml, 'Unable to parse compartment "')

        # Coinciding compartment and species ids
        xml = (
            '<model id="test" '
            'timeUnits="second">'
            '<listOfCompartments>'
            '<compartment id="someId"/>'
            '</listOfCompartments>'
            '<listOfSpecies>'
            '<species id="someId" hasOnlySubstanceUnits="true"'
            ' compartment="someId" constant="false"'
            ' boundaryCondition="true" />'
            '</listOfSpecies>'
            '</model>')
        self.assertBad(xml, 'Unable to parse ')

        # Coinciding parameter and species ids
        xml = (
            '<model id="test" '
            'timeUnits="second">'
            '<listOfCompartments>'
            '<compartment id="someComp"/>'
            '</listOfCompartments>'
            '<listOfParameters>'
            '<parameter id="someId"/>'
            '</listOfParameters>'
            '<listOfSpecies>'
            '<species id="someId" hasOnlySubstanceUnits="true"'
            ' compartment="someComp" constant="false"'
            ' boundaryCondition="true" />'
            '</listOfSpecies>'
            '</model>')
        self.assertBad(xml, 'Unable to parse ')

        # Coinciding parameter and reactant stoichiometry IDs
        stoich_id = 'someStoich'
        xml = (
            '<model id="test" name="test" timeUnits="second">'
            '<listOfCompartments>'
            '<compartment id="someComp"/>'
            '</listOfCompartments>'
            '<listOfParameters>'
            '<parameter id="' + stoich_id + '"/>'
            '</listOfParameters>'
            '<listOfSpecies>'
            '<species id="someSpecies" hasOnlySubstanceUnits="true" '
            'compartment="someComp" constant="false" boundaryCondition="false"'
            '/>'
            '</listOfSpecies>'
            '<listOfReactions>'
            '<reaction id="reaction">'
            '<listOfReactants>'
            '<speciesReference species="someSpecies" '
            'id="' + stoich_id + '"/>'
            '</listOfReactants>'
            '</reaction>'
            '</listOfReactions>'
            '</model>')
        self.assertBad(xml=xml, message='Unable to parse ')

        # Coinciding parameter and product stoichiometry IDs
        stoich_id = 'someStoich'
        xml = (
            '<model id="test" name="test" timeUnits="second">'
            '<listOfCompartments>'
            '<compartment id="someComp"/>'
            '</listOfCompartments>'
            '<listOfParameters>'
            '<parameter id="' + stoich_id + '"/>'
            '</listOfParameters>'
            '<listOfSpecies>'
            '<species id="someSpecies" hasOnlySubstanceUnits="true" '
            'compartment="someComp" constant="false" boundaryCondition="false"'
            '/>'
            '</listOfSpecies>'
            '<listOfReactions>'
            '<reaction id="reaction">'
            '<listOfProducts>'
            '<speciesReference species="someSpecies" '
            'id="' + stoich_id + '"/>'
            '</listOfProducts>'
            '</reaction>'
            '</listOfReactions>'
            '</model>')
        self.assertBad(xml=xml, message='Unable to parse ')

    def test_parse_algebraic_rule(self):
        # Tests parsing algebraic rules (not supported)
        xml = (
            '<model>'
            ' <listOfRules>'
            '  <algebraicRule>'
            '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '    <apply>'
            '    </apply>'
            '   </math>'
            '  </algebraicRule>'
            ' </listOfRules>'
            '</model>'
        )
        self.assertBad(xml, 'Algebraic rules are not supported')

    def test_parse_compartment(self):
        # Test parsing compartments

        a = '<model><listOfCompartments>'
        b = '</listOfCompartments></model>'

        # Test simple compartment
        x = '<compartment id="a" />'
        m = self.parse(a + x + b)
        c = m.compartment('a')
        self.assertEqual(m.compartment('a').sid(), 'a')

        # Missing id
        x = '<compartment/>'
        self.assertBad(a + x + b, 'required attribute "id"')

        # Invalid id
        x = '<compartment id="123" />'
        self.assertBad(a + x + b, 'Invalid SId')

        # Spatial dimensions and model default units
        x = (
            '<model lengthUnits="volt" areaUnits="gram" volumeUnits="lux">'
            ' <listOfCompartments>'
            '  <compartment id="c" />'
            '  <compartment id="s1" spatialDimensions="1" />'
            '  <compartment id="s2" spatialDimensions="2" />'
            '  <compartment id="s3" spatialDimensions="3" />'
            '  <compartment id="d" spatialDimensions="1.4" units="henry" />'
            ' </listOfCompartments>'
            '</model>')
        m = self.parse(x)
        c = m.compartment('c')
        self.assertEqual(c.spatial_dimensions(), None)
        self.assertEqual(c.size_units(), myokit.units.dimensionless)
        c = m.compartment('s1')
        self.assertEqual(c.spatial_dimensions(), 1)
        self.assertEqual(c.size_units(), myokit.units.volt)
        c = m.compartment('s2')
        self.assertEqual(c.spatial_dimensions(), 2)
        self.assertEqual(c.size_units(), myokit.units.g)
        c = m.compartment('s3')
        self.assertEqual(c.spatial_dimensions(), 3)
        self.assertEqual(c.size_units(), myokit.units.lux)
        c = m.compartment('d')
        self.assertEqual(c.spatial_dimensions(), 1.4)
        self.assertEqual(c.size_units(), myokit.units.henry)

        # Unknown units
        x = '<compartment id="x" units="made-up" />'
        self.assertBad(a + x + b, 'Unknown units')

        # Spatial dimensions that cannot be converted to float
        x = '<compartment id="x" spatialDimensions="made-up" />'
        self.assertBad(a + x + b, 'Unable to convert spatial dimensions value')

        # Size that cannot be converted to float
        x = '<compartment id="x" size="made-up" />'
        self.assertBad(a + x + b, 'Unable to convert size')

        # Initial value for size
        x = '<compartment id="x" size="1.32" />'
        m = self.parse(a + x + b)
        c = m.compartment('x')
        self.assertEqual(c.initial_value(), myokit.Number(1.32))

    def test_parse_constraint(self):
        # Test parsing constraints (these are ignored)

        xml = (
            '<model>'
            ' <listOfConstraints>'
            '  <constraint>'
            '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '    <apply> <lt/> <cn> 1 </cn> <ci> S1 </ci> </apply>'
            '   </math>'
            '   <message>'
            '     <p xmlns="http://www.w3.org/1999/xhtml">Out of range.</p>'
            '   </message>'
            '  </constraint>'
            ' </listOfConstraints>'
            '</model>')

        with WarningCollector() as w:
            self.p.parse_string(self.wrap(xml))
        self.assertIn('Ignoring SBML constraints', w.text())

    def test_parse_csymbol_time(self):
        # Test parsing of the csymbol "time"
        xml = (
            '<model timeUnits="second">'
            ' <listOfParameters>'
            '  <parameter id="bacterial_count" value="1" constant="false"/>'
            ' </listOfParameters>'
            ' <listOfRules>'
            '  <assignmentRule variable="bacterial_count">'
            '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '    <apply>'
            '     <times/>'
            '     <ci>bacterial_count</ci>'
            ' <csymbol definitionURL="http://www.sbml.org/sbml/symbols/time"/>'
            '    </apply>'
            '   </math>'
            '  </assignmentRule>'
            ' </listOfRules>'
            '</model>')

        self.p.parse_string(self.wrap(xml))

    def test_parse_events(self):
        # Test parsing events (these are ignored)

        xml = (
            '<model>'
            ' <listOfEvents>'
            '  <event useValuesFromTriggerTime="true">'
            '   <trigger initialValue="false" persistent="true">'
            '    <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '     <apply> <leq/> <ci> P_1 </ci> <ci> P_2 </ci> </apply>'
            '    </math>'
            '   </trigger>'
            '   <listOfEventAssignments>'
            '    <eventAssignment variable="k2">'
            '     <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '      <ci> k2reset </ci>'
            '     </math>'
            '    </eventAssignment>'
            '   </listOfEventAssignments>'
            '  </event>'
            ' </listOfEvents>'
            '</model>')

        with WarningCollector() as w:
            self.p.parse_string(self.wrap(xml))
        self.assertIn('Ignoring SBML events', w.text())

    def test_parse_file(self):
        # Tests parse_file()

        # Supported level
        xml = self.wrap('<model id="a" name="a"/>', 3, 2)
        with WarningCollector() as w:
            self.p.parse_string(xml)
        self.assertNotIn('This version of SBML may not be supported', w.text())

        # Unsupported level
        xml = self.wrap('<model id="a" name="a"/>', 2, 2)
        with WarningCollector() as w:
            self.p.parse_string(xml)
        self.assertIn('This version of SBML may not be supported', w.text())

        # Unsupported version
        xml = self.wrap('<model id="a" name="a"/>', 3, 1)
        with WarningCollector() as w:
            self.p.parse_string(xml)
        self.assertIn('This version of SBML may not be supported', w.text())

        # Check whether error is thrown for invalid path
        path = 'some/path'
        message = 'Unable to parse XML: '
        self.assertRaisesRegex(
            SBMLParsingError,
            message,
            self.p.parse_file,
            path)

        # Check whether error is thrown for invalid xml
        self.assertBad('<model>', 'Unable to parse XML: ')

    def test_function_definitions(self):
        # Function definitions are not supported

        xml = (
            '<model id="test" name="test" timeUnits="second"> '
            ' <listOfFunctionDefinitions>'
            '  <functionDefinition id="multiply" name="multiply">'
            '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
            '    <lambda>'
            '     <bvar>'
            '      <ci> x </ci>'
            '     </bvar>'
            '     <bvar>'
            '      <ci> y </ci>'
            '     </bvar>'
            '     <apply>'
            '      <times/>'
            '       <ci> x </ci>'
            '       <ci> y </ci>'
            '     </apply>'
            '    </lambda>'
            '   </math>'
            '  </functionDefinition>'
            ' </listOfFunctionDefinitions>'
            '</model>')
        self.assertBad(xml, 'Function definitions are not supported.')

    def test_parse_initial_assignment(self):
        # Test parsing initial assignments.

        # Set a parameter value
        a = ('<model name="mathml">'
             ' <listOfParameters>'
             '  <parameter id="x" constant="true" />'
             ' </listOfParameters>'
             ' <listOfInitialAssignments>')
        b = ('  <initialAssignment symbol="x">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <cn>5</cn>'
             '   </math>'
             '  </initialAssignment>')
        c = (' </listOfInitialAssignments>'
             '</model>')
        p = self.parse(a + b + c).parameter('x')
        self.assertEqual(p.initial_value(), myokit.Number(5))

        # Reset a parameter value
        x = ('<model name="mathml">'
             ' <listOfParameters>'
             '  <parameter id="x" value="3" constant="true" />'
             ' </listOfParameters>'
             ' <listOfInitialAssignments>')
        p = self.parse(x + c).parameter('x')
        self.assertEqual(p.initial_value(), myokit.Number(3))
        p = self.parse(x + b + c).parameter('x')
        self.assertEqual(p.initial_value(), myokit.Number(5))

        # No maths: warning
        y = '<initialAssignment symbol="x"></initialAssignment>'
        with WarningCollector() as w:
            p = self.parse(x + y + c).parameter('x')
        self.assertIn('nitial assignment does not define any math', w.text())
        self.assertEqual(p.initial_value(), myokit.Number(3))

        # Invalid maths: error
        x = ('  <initialAssignment symbol="x">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <apply><power /><cn>1</cn></apply>'
             '   </math>'
             '  </initialAssignment>')
        self.assertBad(a + x + c, r'Expecting 2 operand\(s\), got 1 for power')

        # Missing symbol: error
        x = '<initialAssignment />'
        self.assertBad(a + x + c, 'required attribute "symbol"')

        # Invalid/unknown symbol
        x = ('  <initialAssignment symbol="blue">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <cn>5</cn>'
             '   </math>'
             '  </initialAssignment>')
        self.assertBad(a + x + c, 'SId "blue" does not refer to')

        # Set a compartment size
        a = ('<model name="mathml">')
        b = (' <listOfCompartments>'
             '  <compartment id="c" />'
             ' </listOfCompartments>')
        c = (' <listOfInitialAssignments>'
             '  <initialAssignment symbol="c">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <cn>1.23</cn>'
             '   </math>'
             '  </initialAssignment>'
             ' </listOfInitialAssignments>')
        d = ('</model>')
        m = self.parse(a + b + c + d)
        self.assertEqual(
            m.compartment('c').initial_value(), myokit.Number((1.23)))

        # Reset a compartment size
        b = ('<listOfCompartments>'
             '<compartment id="c" size="12" />'
             '</listOfCompartments>')
        m = self.parse(a + b + d)
        self.assertEqual(
            m.compartment('c').initial_value(), myokit.Number((12)))
        m = self.parse(a + b + c + d)
        self.assertEqual(
            m.compartment('c').initial_value(), myokit.Number((1.23)))

        # Set a species amount (or concentration)
        a = ('<model>'
             ' <listOfCompartments>'
             '  <compartment id="c" size="1.2" />'
             ' </listOfCompartments>')
        b = (' <listOfSpecies>'
             '  <species id="s" compartment="c" />'
             ' </listOfSpecies>')
        c = (' <listOfInitialAssignments>'
             '  <initialAssignment symbol="s">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <cn>4.5</cn>'
             '   </math>'
             '  </initialAssignment>'
             ' </listOfInitialAssignments>')
        d = ('</model>')
        m = self.parse(a + b + c + d)
        s = m.species('s')
        expr, expr_in_amount = s.initial_value()
        self.assertEqual(expr, myokit.Number(4.5))
        self.assertIsNone(expr_in_amount)

        # Reset a species amount (or concentration)
        b = (' <listOfSpecies>'
             '  <species id="s" compartment="c" initialConcentration="3" />'
             ' </listOfSpecies>')
        m = self.parse(a + b + d)
        s = m.species('s')
        expr, expr_in_amount = s.initial_value()
        self.assertEqual(expr, myokit.Number(3))
        self.assertFalse(expr_in_amount)
        m = self.parse(a + b + c + d)
        s = m.species('s')
        expr, expr_in_amount = s.initial_value()
        self.assertEqual(expr, myokit.Number(4.5))
        self.assertIsNone(expr_in_amount)

        # Set a stoichiometry
        a = ('<model>'
             ' <listOfCompartments>'
             '  <compartment id="c" size="1.2" />'
             ' </listOfCompartments>'
             ' <listOfSpecies>'
             '  <species id="s" compartment="c" />'
             ' </listOfSpecies>'
             ' <listOfReactions>'
             '  <reaction id="r">'
             '   <listOfReactants>')
        b = ('    <speciesReference species="s" id="sr" />')
        c = ('   </listOfReactants>'
             '   <kineticLaw>'
             '    <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <apply><ci>s</ci></apply>'
             '    </math>'
             '   </kineticLaw>'
             '  </reaction>'
             ' </listOfReactions>')
        d = (' <listOfInitialAssignments>'
             '  <initialAssignment symbol="sr">'
             '   <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <cn>4.51</cn>'
             '   </math>'
             '  </initialAssignment>'
             ' </listOfInitialAssignments>')
        e = ('</model>')
        sr = self.parse(a + b + c + d + e).reaction('r').reactants()[0]
        self.assertEqual(sr.initial_value(), myokit.Number(4.51))

        # Reset a stoichiometry
        b = '<speciesReference species="s" id="sr" stoichiometry="2" />'
        sr = self.parse(a + b + c + e).reaction('r').reactants()[0]
        self.assertEqual(sr.initial_value(), myokit.Number(2))
        sr = self.parse(a + b + c + d + e).reaction('r').reactants()[0]
        self.assertEqual(sr.initial_value(), myokit.Number(4.51))

    def test_parse_model(self):
        # Tests parsing a model.

        # No model
        self.assertBad(xml='', message='Model element not found.')
        self.assertBad(xml='<hello/>', message='Model element not found.')

        # Name from name attribute
        m = self.parse('<model id="yes" name="no" />')
        self.assertEqual(m.name(), 'no')

        # Name from id attribute
        m = self.parse('<model id="yes" />')
        self.assertEqual(m.name(), 'yes')

        # Default name
        m = self.parse('<model />')
        self.assertEqual(m.name(), 'Imported SBML model')

        # Notes
        m = self.parse(
            '<model>'
            ' <notes>'
            '  <body xmlns="http://www.w3.org/1999/xhtml">'
            '   <h1>This is a bunch of notes</h1>'
            '   <div><p>With sentences like this one</p></div>'
            '   <p>And a<br />line break</p>'
            '  </body>'
            ' </notes>'
            '</model>'
        )
        self.assertIn('sentences like this one', m.notes())

        # Global units: length, area, volume, substance, extent, time
        m = self.parse('<model />')
        self.assertEqual(m.length_units(), myokit.units.dimensionless)
        self.assertEqual(m.area_units(), myokit.units.dimensionless)
        self.assertEqual(m.volume_units(), myokit.units.dimensionless)
        self.assertEqual(m.substance_units(), myokit.units.dimensionless)
        self.assertEqual(m.extent_units(), myokit.units.dimensionless)
        self.assertEqual(m.time_units(), myokit.units.dimensionless)

        m = self.parse(
            '<model'
            ' lengthUnits="meter"'
            ' areaUnits="meter_squared"'
            ' volumeUnits="candela"'
            ' substanceUnits="coulomb"'
            ' extentUnits="hertz"'
            ' timeUnits="second"'
            '>'
            ' <listOfUnitDefinitions>'
            '  <unitDefinition id="meter_squared">'
            '   <listOfUnits>'
            '    <unit kind="meter" exponent="2" />'
            '   </listOfUnits>'
            '  </unitDefinition>'
            ' </listOfUnitDefinitions>'
            '</model>'
        )
        self.assertEqual(m.length_units(), myokit.units.m)
        self.assertEqual(m.area_units(), myokit.units.m**2)
        self.assertEqual(m.volume_units(), myokit.units.candela)
        self.assertEqual(m.substance_units(), myokit.units.coulomb)
        self.assertEqual(m.extent_units(), myokit.units.hertz)
        self.assertEqual(m.time_units(), myokit.units.s)

        self.assertBad(
            '<model areaUnits="meter_squared" />',
            'Error parsing model element: ')
        self.assertBad(
            '<model areaUnits="celsius" />',
            'Error parsing model element: ')

        # Global conversion factor
        m = self.parse('<model />')
        self.assertIsNone(m.conversion_factor())
        m = self.parse(
            '<model conversionFactor="x">'
            ' <listOfParameters>'
            '  <parameter id="x" />'
            ' </listOfParameters>'
            '</model>'
        )
        self.assertEqual(m.conversion_factor(), m.parameter('x'))

        self.assertBad(
            '<model conversionFactor="x" />', 'Model conversion factor')

    def test_parse_parameter(self):
        # Tests parsing parameters

        a = '<model><listOfParameters>'
        b = '</listOfParameters></model>'

        # Named
        x = '<parameter id="a" /><parameter id="b" />'
        m = self.parse(a + x + b)
        self.assertEqual(m.parameter('a').sid(), 'a')
        self.assertEqual(m.parameter('b').sid(), 'b')

        # With value and units
        x = ('<parameter id="c" value="2" />'
             '<parameter id="d" units="volt" />'
             '<parameter id="e" units="ampere" value="-1.2e-3" />')
        m = self.parse(a + x + b)
        c = m.parameter('c')
        d = m.parameter('d')
        e = m.parameter('e')
        self.assertEqual(c.sid(), 'c')
        self.assertIsNone(c.units())
        self.assertIsInstance(c.initial_value(), myokit.Number)
        self.assertEqual(c.initial_value().eval(), 2)
        self.assertEqual(d.sid(), 'd')
        self.assertEqual(d.units(), myokit.units.volt)
        self.assertIsNone(d.initial_value())
        self.assertEqual(e.sid(), 'e')
        self.assertEqual(e.units(), myokit.units.ampere)
        self.assertEqual(e.initial_value().eval(), -1.2e-3)
        self.assertIsInstance(e.initial_value(), myokit.Number)

        # Missing id
        x = '<parameter/>'
        self.assertBad(a + x + b, 'required attribute "id"')

        # Invalid id
        x = '<parameter id="123" />'
        self.assertBad(a + x + b, 'Invalid SId')

        # Bad units
        x = '<parameter id="x" units="made-up" />'
        self.assertBad(a + x + b, 'Unknown units')

        # Bad value
        x = '<parameter id="x" value="made-up" />'
        self.assertBad(a + x + b, 'Unable to convert parameter value')

    def test_parse_reaction(self):
        # Test parsing reactions, species references, kinetic laws

        a = ('<model>'
             ' <listOfCompartments>'
             '  <compartment id="c" size="1.2" />'
             ' </listOfCompartments>'
             ' <listOfSpecies>'
             '  <species id="S1" compartment="c" initialAmount="0"/>'
             '  <species id="S2" compartment="c" initialAmount="0"/>'
             '  <species id="S3" compartment="c" initialAmount="0"/>'
             '  <species id="S4" compartment="c" initialAmount="0"/>'
             ' </listOfSpecies>'
             ' <listOfParameters>'
             '  <parameter id="k1" value="3" />'
             ' </listOfParameters>'
             ' <listOfReactions>')
        b = ('  <reaction id="r">'
             '   <listOfReactants>'
             '    <speciesReference species="S1" stoichiometry="2" id="r1" />'
             '   </listOfReactants>'
             '   <listOfProducts>'
             '    <speciesReference species="S2" stoichiometry="5" id="r2" />'
             '   </listOfProducts>'
             '   <listOfModifiers>'
             '    <modifierSpeciesReference species="S3" id="r3" />'
             '   </listOfModifiers>')
        c = ('   <kineticLaw>'
             '    <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <apply>'
             '      <times/>'
             '      <ci> c </ci>'
             '      <ci> k1 </ci>'
             '      <ci> S1 </ci>'
             '     </apply>'
             '    </math>'
             '   </kineticLaw>')
        d = ('  </reaction>')
        e = (' </listOfReactions>'
             '</model>')

        # Parse valid reaction
        m = self.parse(a + b + c + d + e)
        r = m.reaction('r')
        self.assertEqual(r.sid(), 'r')

        # Check reaction knows S1, S2, S3 but not S4
        self.assertEqual(r.species('S1'), m.species('S1'))
        self.assertEqual(r.species('S2'), m.species('S2'))
        self.assertEqual(r.species('S3'), m.species('S3'))
        self.assertRaises(KeyError, r.species, 'S4')
        m.species('S4')

        # Check reactants, products, modifiers
        s1r = r.reactants()
        self.assertEqual(len(s1r), 1)
        s1r = s1r[0]
        self.assertEqual(s1r.species(), m.species('S1'))
        self.assertEqual(s1r.initial_value(), myokit.Number(2))
        s2r = r.products()
        self.assertEqual(len(s2r), 1)
        s2r = s2r[0]
        self.assertEqual(s2r.species(), m.species('S2'))
        self.assertEqual(s2r.initial_value(), myokit.Number(5))
        s3r = r.modifiers()
        self.assertEqual(len(s3r), 1)
        s3r = s3r[0]
        self.assertEqual(s3r.species(), m.species('S3'))
        with self.assertRaises(AttributeError):
            s3r.initial_value()

        # Reactants and products are assignable, modifier is not
        self.assertEqual(m.assignable('r1'), s1r)
        self.assertEqual(m.assignable('r2'), s2r)
        self.assertRaises(KeyError, m.assignable, 'r3')

        # Check kinetic law
        self.assertEqual(
            r.kinetic_law().code(),
            '<Compartment c> * <Parameter k1> * <Species S1>')

        # Fast=true raises an error
        self.assertBad(
            a + '<reaction id="r" fast="true" />' + e, 'not supported')

        # Reactions must have a valid id
        self.assertBad(a + '<reaction />' + e, 'required attribute "id"')
        self.assertBad(a + '<reaction id="" />' + e, 'Invalid SId')

        # Reactions must have a product or reactant
        x = ('<reaction id="r" />')
        self.assertBad(a + x + e, 'at least one reactant or product')

        # Missing or empty kinetic law is ignored
        with WarningCollector() as w:
            self.parse(a + b + d + e)
        self.assertEqual(w.count(), 1)
        self.assertIn('No kinetic law set for reaction "r"', w.text())
        x = '<kineticLaw />'
        with WarningCollector() as w:
            m = self.parse(a + b + x + d + e)
        self.assertEqual(w.count(), 1)
        self.assertIn('No kinetic law set for reaction "r"', w.text())

        # Kinetic law: using species not used in reaction
        x = ('<kineticLaw>'
             ' <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '  <apply><ci>S4</ci></apply>'
             ' </math>'
             '</kineticLaw>')
        self.assertBad(a + b + x + d + e, 'Unknown or inaccessible')

        # Kinetic law: local parameters are not supported
        x = ('<kineticLaw>'
             ' <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '  <apply><ci>k</ci></apply>'
             ' </math>'
             ' <listOfLocalParameters>'
             '  <localParameter id="k" value="3.4" />'
             ' </listOfLocalParameters>'
             '</kineticLaw>')
        self.assertBad(a + b + x + d + e, 'Local parameters')

        # Kinetic law: invalid MathML
        x = ('<kineticLaw>'
             ' <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '  <apply><species>S3</species></apply>'
             ' </math>'
             '</kineticLaw>')
        self.assertBad(a + b + x + d + e, 'Unsupported element')

        # Unknown species in species reference
        x = ('<reaction id="r">'
             ' <listOfReactants>'
             '  <speciesReference species="S1" stoichiometry="2" id="r1" />'
             '  <speciesReference species="S5" stoichiometry="5" id="r2" />'
             ' </listOfReactants>')
        self.assertBad(a + x + c + d + e, 'unknown species "S5"')

        # Invalid stoichiometry
        x = ('<reaction id="r">'
             ' <listOfReactants>'
             '  <speciesReference species="S1" stoichiometry="2" id="r1" />'
             '  <speciesReference species="S2" stoichiometry="one" id="r2" />'
             ' </listOfReactants>')
        self.assertBad(a + x + c + d + e, 'Unable to convert stoichiometry')

        # Check that rate of change equations are correctly computed from
        # multiple reactions
        f = ('  <reaction id="reaction2">'
             '   <listOfReactants>'
             '    <speciesReference species="S1" />'
             '   </listOfReactants>'
             '   <listOfProducts>'
             '    <speciesReference species="S3" />'
             '   </listOfProducts>'
             '   <listOfModifiers>'
             '    <modifierSpeciesReference species="S4" />'
             '   </listOfModifiers>'
             '   <kineticLaw>'
             '    <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <apply>'
             '      <times/>'
             '      <ci> c </ci>'
             '      <ci> k1 </ci>'
             '      <ci> S1 </ci>'
             '     </apply>'
             '    </math>'
             '   </kineticLaw>'
             '  </reaction>')

        # Parse reaction
        m = self.parse(a + b + c + d + f + e)

        # Check first reaction
        r = m.reaction('r')
        self.assertEqual(r.sid(), 'r')

        # Check reaction knows S1, S2, S3 but not S4
        self.assertEqual(r.species('S1'), m.species('S1'))
        self.assertEqual(r.species('S2'), m.species('S2'))
        self.assertEqual(r.species('S3'), m.species('S3'))
        self.assertRaises(KeyError, r.species, 'S4')
        m.species('S4')

        # Check reactants, products, modifiers
        s1r = r.reactants()
        self.assertEqual(len(s1r), 1)
        s1r = s1r[0]
        self.assertEqual(s1r.species(), m.species('S1'))
        self.assertEqual(s1r.initial_value(), myokit.Number(2))
        s2r = r.products()
        self.assertEqual(len(s2r), 1)
        s2r = s2r[0]
        self.assertEqual(s2r.species(), m.species('S2'))
        self.assertEqual(s2r.initial_value(), myokit.Number(5))
        s3r = r.modifiers()
        self.assertEqual(len(s3r), 1)
        s3r = s3r[0]
        self.assertEqual(s3r.species(), m.species('S3'))
        with self.assertRaises(AttributeError):
            s3r.initial_value()

        # Reactants and products are assignable, modifier is not
        self.assertEqual(m.assignable('r1'), s1r)
        self.assertEqual(m.assignable('r2'), s2r)
        self.assertRaises(KeyError, m.assignable, 'r3')

        # Check kinetic law
        self.assertEqual(
            r.kinetic_law().code(),
            '<Compartment c> * <Parameter k1> * <Species S1>')

        # Check second reaction
        r = m.reaction('reaction2')
        self.assertEqual(r.sid(), 'reaction2')

        # Check reaction knows S1, S3, S4 but not S2
        self.assertEqual(r.species('S1'), m.species('S1'))
        self.assertEqual(r.species('S3'), m.species('S3'))
        self.assertEqual(r.species('S4'), m.species('S4'))
        self.assertRaises(KeyError, r.species, 'S2')
        m.species('S2')

        # Check kinetic law
        self.assertEqual(
            r.kinetic_law().code(),
            '<Compartment c> * <Parameter k1> * <Species S1>')

        # Check species reactions (no unecessary zeros and all terms present)
        model_mmt = m.myokit_model().code()
        S1_rhs = \
            'dot(S1_amount) = -(r1 * (size * global.k1 * S1_concentration)) ' \
            + '- size * global.k1 * S1_concentration'
        self.assertTrue(S1_rhs in model_mmt)
        S2_rhs = 'dot(S2_amount) = r2 * (size * global.k1 * S1_concentration)'
        self.assertTrue(S2_rhs in model_mmt)
        S3_rhs = 'dot(S3_amount) = size * global.k1 * S1_concentration'
        self.assertTrue(S3_rhs in model_mmt)

    def test_parse_rule(self):
        # Tests parsing assignment rules and rate rules

        # Set a parameter value
        a = ('<model name="mathml">'
             ' <listOfParameters>'
             '  <parameter id="x" constant="true" value="3"/>'
             '  <parameter id="y" constant="true" value="5"/>'
             ' </listOfParameters>')
        b = ('</model>')

        p = self.parse(a + b).parameter('x')
        self.assertEqual(p.initial_value(), myokit.Number(3))

        p = self.parse(a + b).parameter('y')
        self.assertEqual(p.initial_value(), myokit.Number(5))

        # Set RHS with assignmentRule or rateRule
        x = ('<listOfRules>'
             '  <assignmentRule variable="x">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> y </ci>'
             '      <cn> 3 </cn>'
             '    </apply>'
             '  </math>'
             '  </assignmentRule>'
             '  <rateRule variable="y">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> y </ci>'
             '      <apply>'
             '        <power/>'
             '        <ci> x </ci>'
             '        <cn> 2 </cn>'
             '      </apply>'
             '    </apply>'
             '  </math>'
             '  </rateRule>'
             '</listOfRules>')

        px = self.parse(a + x + b).parameter('x')
        py = self.parse(a + x + b).parameter('y')

        self.assertEqual(px.initial_value(), myokit.Number(3))
        self.assertEqual(py.initial_value(), myokit.Number(5))

        self.assertFalse(px.is_rate())
        self.assertTrue(py.is_rate())

        expr = myokit.Multiply(myokit.Name(py), myokit.Number(3))
        self.assertEqual(px.value().code(), expr.code())

        expr = myokit.Multiply(
            myokit.Name(py),
            myokit.Power(myokit.Name(px), myokit.Number(2)))
        self.assertEqual(py.value().code(), expr.code())

        # Set a compartment size
        a = ('<model name="mathml">'
             ' <listOfCompartments>'
             '  <compartment id="cx" size="10"/>'
             '  <compartment id="cy" size="5"/>'
             ' </listOfCompartments>')
        b = ('</model>')

        p = self.parse(a + b).compartment('cx')
        self.assertEqual(p.initial_value(), myokit.Number(10))

        p = self.parse(a + b).compartment('cy')
        self.assertEqual(p.initial_value(), myokit.Number(5))

        # Set RHS with assignmentRule or rateRule
        x = ('<listOfRules>'
             '  <assignmentRule variable="cx">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> cy </ci>'
             '      <cn> 3 </cn>'
             '    </apply>'
             '  </math>'
             '  </assignmentRule>'
             '  <rateRule variable="cy">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> cy </ci>'
             '      <apply>'
             '        <power/>'
             '        <ci> cx </ci>'
             '        <cn> 2 </cn>'
             '      </apply>'
             '    </apply>'
             '  </math>'
             '  </rateRule>'
             '</listOfRules>')

        px = self.parse(a + x + b).compartment('cx')
        py = self.parse(a + x + b).compartment('cy')

        self.assertEqual(px.initial_value(), myokit.Number(10))
        self.assertEqual(py.initial_value(), myokit.Number(5))

        self.assertFalse(px.is_rate())
        self.assertTrue(py.is_rate())

        expr = myokit.Multiply(myokit.Name(py), myokit.Number(3))
        self.assertEqual(px.value().code(), expr.code())

        expr = myokit.Multiply(
            myokit.Name(py),
            myokit.Power(myokit.Name(px), myokit.Number(2)))
        self.assertEqual(py.value().code(), expr.code())

        # Set a species amount (or concentration)
        a = ('<model name="mathml">'
             ' <listOfCompartments>'
             '  <compartment id="c" size="10"/>'
             ' </listOfCompartments>'
             ' <listOfSpecies>'
             '  <species id="sx" compartment="c" initialAmount="3.4"'
             '    boundaryCondition="true"/>'
             '  <species id="sy" compartment="c" initialConcentration="1.2"'
             '    boundaryCondition="true"/>'
             ' </listOfSpecies>')
        b = ('</model>')

        p = self.parse(a + b).species('sx')
        expr, expr_in_amount = p.initial_value()
        self.assertEqual(expr, myokit.Number(3.4))
        self.assertTrue(expr_in_amount)

        p = self.parse(a + b).species('sy')
        expr, expr_in_amount = p.initial_value()
        self.assertEqual(expr, myokit.Number(1.2))
        self.assertFalse(expr_in_amount)

        # Set RHS with assignmentRule or rateRule
        x = ('<listOfRules>'
             '  <assignmentRule variable="sx">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> sy </ci>'
             '      <cn> 3 </cn>'
             '    </apply>'
             '  </math>'
             '  </assignmentRule>'
             '  <rateRule variable="sy">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> sy </ci>'
             '      <apply>'
             '        <power/>'
             '        <ci> sx </ci>'
             '        <cn> 2 </cn>'
             '      </apply>'
             '    </apply>'
             '  </math>'
             '  </rateRule>'
             '</listOfRules>')

        px = self.parse(a + x + b).species('sx')
        py = self.parse(a + x + b).species('sy')

        expr, expr_in_amount = px.initial_value()
        self.assertEqual(expr, myokit.Number(3.4))
        self.assertTrue(expr_in_amount)

        expr, expr_in_amount = py.initial_value()
        self.assertEqual(expr, myokit.Number(1.2))
        self.assertFalse(expr_in_amount)

        self.assertFalse(px.is_rate())
        self.assertTrue(py.is_rate())

        expr = myokit.Multiply(myokit.Name(py), myokit.Number(3))
        self.assertEqual(px.value().code(), expr.code())

        expr = myokit.Multiply(
            myokit.Name(py),
            myokit.Power(myokit.Name(px), myokit.Number(2)))
        self.assertEqual(py.value().code(), expr.code())

        # Non-boundary species
        y = ('<model name="mathml">'
             ' <listOfCompartments>'
             '  <compartment id="c" size="10"/>'
             ' </listOfCompartments>'
             ' <listOfSpecies>'
             '  <species id="sx" compartment="c" initialAmount="3.4"/>'
             '  <species id="sy" compartment="c" initialConcentration="1.2"/>'
             ' </listOfSpecies>')

        self.assertBad(
            y + x + b, 'Assignment or rate rule set for species that is')

        # Set a stoichiometry
        a = ('<model name="mathml">'
             ' <listOfCompartments>'
             '  <compartment id="c" size="10"/>'
             ' </listOfCompartments>'
             ' <listOfSpecies>'
             '  <species id="s1" compartment="c" />'
             '  <species id="s2" compartment="c" />'
             ' </listOfSpecies>'
             ' <listOfReactions>'
             '  <reaction id="r">'
             '   <listOfReactants>'
             '    <speciesReference species="s1" id="sx" stoichiometry="2.1"/>'
             '   </listOfReactants>'
             '   <listOfProducts>'
             '    <speciesReference species="s2" id="sy" stoichiometry="3.5"/>'
             '   </listOfProducts>'
             '   <kineticLaw>'
             '    <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '     <apply>'
             '      <plus/>'
             '      <ci>s1</ci>'
             '      <ci>s2</ci>'
             '     </apply>'
             '    </math>'
             '   </kineticLaw>'
             '  </reaction>'
             ' </listOfReactions>')
        b = ('</model>')

        reaction = self.parse(a + x + b).reaction('r')
        px = reaction.reactants().pop()
        py = reaction.products().pop()

        self.assertEqual(px.initial_value(), myokit.Number(2.1))
        self.assertEqual(py.initial_value(), myokit.Number(3.5))

        # Set RHS with assignmentRule or rateRule
        x = ('<listOfRules>'
             '  <assignmentRule variable="sx">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> sy </ci>'
             '      <cn> 3 </cn>'
             '    </apply>'
             '  </math>'
             '  </assignmentRule>'
             '  <rateRule variable="sy">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <ci> sy </ci>'
             '      <apply>'
             '        <power/>'
             '        <ci> sx </ci>'
             '        <cn> 2 </cn>'
             '      </apply>'
             '    </apply>'
             '  </math>'
             '  </rateRule>'
             '</listOfRules>')

        reaction = self.parse(a + x + b).reaction('r')
        px = reaction.reactants().pop()
        py = reaction.products().pop()

        self.assertEqual(px.initial_value(), myokit.Number(2.1))
        self.assertEqual(py.initial_value(), myokit.Number(3.5))

        self.assertFalse(px.is_rate())
        self.assertTrue(py.is_rate())

        expr = myokit.Multiply(myokit.Name(py), myokit.Number(3))
        self.assertEqual(px.value().code(), expr.code())

        expr = myokit.Multiply(
            myokit.Name(py),
            myokit.Power(myokit.Name(px), myokit.Number(2)))
        self.assertEqual(py.value().code(), expr.code())

        # No maths: warning
        a = ('<model name="mathml">'
             ' <listOfParameters>'
             '  <parameter id="x" constant="true" value="3"/>'
             '  <parameter id="y" constant="true" value="5"/>'
             ' </listOfParameters>')
        b = ('</model>')

        x = ('<listOfRules>'
             '  <assignmentRule variable="x">'
             '  </assignmentRule>'
             '</listOfRules>')
        with WarningCollector() as w:
            p = self.parse(a + x + b).parameter('x')
        self.assertIn('Rule does not define any mathematics', w.text())

        x = ('<listOfRules>'
             '  <rateRule variable="y">'
             '  </rateRule>'
             '</listOfRules>')
        with WarningCollector() as w:
            p = self.parse(a + x + b).parameter('y')
        self.assertIn('Rule does not define any mathematics', w.text())

        # Bad maths: error
        x = ('<listOfRules>'
             '  <assignmentRule variable="x">'
             '  <math xmlns="http://www.w3.org/1998/Math/MathML">'
             '    <apply>'
             '      <times/>'
             '      <weird_maths> 2 </weird_maths>'
             '      <cn> 3 </cn>'
             '    </apply>'
             '  </math>'
             '  </assignmentRule>'
             '</listOfRules>')
        self.assertBad(a + x + b, 'Unable to parse rule: ')

        # Missing symbol: error
        x = ('<listOfRules>'
             '  <rateRule>'
             '  </rateRule>'
             '</listOfRules>')
        self.assertBad(a + x + b, 'Element')

        # Invalid/unknown symbol
        x = ('<listOfRules>'
             '  <rateRule variable="blue">'
             '  </rateRule>'
             '</listOfRules>')
        self.assertBad(a + x + b, 'SId "blue" does not refer to')

    def test_parse_species(self):
        # Tests parsing species

        a = ('<model>'
             ' <listOfCompartments>'
             '  <compartment id="c" />'
             ' </listOfCompartments>'
             ' <listOfSpecies>')
        b = (' </listOfSpecies>'
             ' <listOfParameters>'
             '  <parameter id="p" />'
             ' </listOfParameters>'
             '</model>')

        # Simple species
        x = '<species compartment="c" id="spec" />'
        s0 = self.parse(a + x + b).species('spec')
        self.assertTrue(s0.sid() == 'spec')

        # Missing id
        x = '<species compartment="c" />'
        self.assertBad(a + x + b, 'required attribute "id"')

        # Invalid id
        x = '<species compartment="c" id="456" />'
        self.assertBad(a + x + b, 'Invalid SId')

        # Missing compartment
        x = '<species id="spec" />'
        self.assertBad(a + x + b, 'required attribute "compartment"')

        # Unknown compartment
        x = '<species id="spec" compartment="hello" />'
        self.assertBad(a + x + b, 'Unknown compartment')

        # Amount or concentration
        self.assertFalse(s0.is_amount())
        x = '<species compartment="c" id="s" hasOnlySubstanceUnits="false" />'
        s = self.parse(a + x + b).species('s')
        self.assertFalse(s.is_amount())
        x = '<species compartment="c" id="s" hasOnlySubstanceUnits="true" />'
        s = self.parse(a + x + b).species('s')
        self.assertTrue(s.is_amount())

        # Boundary
        self.assertFalse(s0.is_boundary())
        x = '<species compartment="c" id="s" boundaryCondition="false" />'
        s = self.parse(a + x + b).species('s')
        self.assertFalse(s.is_boundary())
        x = '<species compartment="c" id="s" boundaryCondition="true" />'
        s = self.parse(a + x + b).species('s')
        self.assertTrue(s.is_boundary())

        # Constant
        self.assertFalse(s0.is_constant())
        x = '<species compartment="c" id="s" constant="false" />'
        s = self.parse(a + x + b).species('s')
        self.assertFalse(s.is_constant())
        x = '<species compartment="c" id="s" constant="true" />'
        s = self.parse(a + x + b).species('s')
        self.assertTrue(s.is_constant())

        # Substance units
        self.assertEqual(s0.substance_units(), myokit.units.dimensionless)
        x = '<species compartment="c" id="s" substanceUnits="volt" />'
        s = self.parse(a + x + b).species('s')
        self.assertEqual(s.substance_units(), myokit.units.volt)
        x = ('<species compartment="c" id="s"'
             ' substanceUnits="volt" hasOnlySubstanceUnits="true" />')
        s = self.parse(a + x + b).species('s')
        self.assertEqual(s.substance_units(), myokit.units.volt)

        # Invalid units
        x = '<species compartment="c" id="s" substanceUnits="made-up" />'
        self.assertBad(a + x + b, 'Unknown units')

        # Initial amount
        x = ('<species compartment="c" id="s"'
             ' hasOnlySubstanceUnits="true" initialAmount="3" />')
        s = self.parse(a + x + b).species('s')
        expr, expr_in_amount = s.initial_value()
        self.assertEqual(expr, myokit.Number(3))
        self.assertTrue(expr_in_amount)

        # Initial concentration
        x = ('<species compartment="c" id="s"'
             ' hasOnlySubstanceUnits="false" initialConcentration="1.2" />')
        s = self.parse(a + x + b).species('s')
        expr, expr_in_amount = s.initial_value()
        self.assertEqual(expr, myokit.Number(1.2))
        self.assertFalse(expr_in_amount)

        # Set both initial amount and concentration
        x = ('<species compartment="c" id="s"'
             ' hasOnlySubstanceUnits="true" initialAmount="3"'
             ' initialConcentration="10"/>')
        self.assertBad(
            a + x + b, 'Species cannot set both an initialAmount and an')

        # Invalid initial value
        x = ('<species compartment="c" id="s"'
             ' hasOnlySubstanceUnits="true" initialAmount="made-up" />')
        self.assertBad(
            a + x + b, 'Unable to convert initial species value to float "')

        # Conversion factor parameter
        x = ('<species compartment="c" id="s" conversionFactor="p" />')
        m = self.parse(a + x + b)
        s = m.species('s')
        self.assertEqual(s.conversion_factor(), m.parameter('p'))

        # Conversion factor points to unknown parameter
        x = ('<species compartment="c" id="s" conversionFactor="q" />')
        self.assertBad(a + x + b, 'Unknown parameter')

    def test_parse_species_reference(self):
        # TODO:
        pass

    def test_parse_unit(self):
        # Tests parsing units and unit definitions

        a = '<model><listOfUnitDefinitions>'
        b = '</listOfUnitDefinitions></model>'

        # Dimensionless units
        xml = '<unitDefinition id="dimless" />'
        m = self.parse(a + xml + b)
        self.assertEqual(m.unit('dimless'), myokit.units.dimensionless)

        # Dimensionless units
        xml = (
            '<unitDefinition id="dimless">'
            ' <listOfUnits />'
            '</unitDefinition>')
        m = self.parse(a + xml + b)
        self.assertEqual(m.unit('dimless'), myokit.units.dimensionless)

        # Full featured units
        xml = (
            '<unitDefinition id="centimeter">'
            ' <listOfUnits>'
            '  <unit kind="meter" scale="-2" />'
            ' </listOfUnits>'
            '</unitDefinition>'
            '<unitDefinition id="supervolt">'
            ' <listOfUnits>'
            '  <unit kind="volt" multiplier="2.3" scale="3" exponent="2" />'
            '  <unit kind="meter" exponent="-1" />'
            ' </listOfUnits>'
            '</unitDefinition>')
        m = self.parse(a + xml + b)
        self.assertEqual(m.unit('centimeter'), myokit.units.cm)
        sv = (myokit.units.volt * 2.3e3)**2 / myokit.units.meter
        self.assertEqual(m.unit('supervolt'), sv)

        # Parameter with custom unit
        x = ('<model>'
             ' <listOfUnitDefinitions>'
             '  <unitDefinition id="centimeter">'
             '   <listOfUnits>'
             '    <unit kind="meter" scale="-2" />'
             '   </listOfUnits>'
             '  </unitDefinition>'
             ' </listOfUnitDefinitions>'
             ' <listOfParameters>'
             '  <parameter id="a" units="centimeter" />'
             ' </listOfParameters>'
             '</model>')
        m = self.parse(x)
        self.assertEqual(m.parameter('a').units(), myokit.units.cm)

        # Missing id
        xml = '<unitDefinition />'
        self.assertBad(a + xml + b, 'required attribute "id"')

        # Invalid id
        xml = '<unitDefinition id="123" />'
        self.assertBad(a + xml + b, 'Invalid UnitSId')

        # Invalid unit kind
        xml = (
            '<unitDefinition id="centimeter">'
            ' <listOfUnits>'
            '  <unit kind="celsius" scale="-2" />'
            ' </listOfUnits>'
            '</unitDefinition>')
        self.assertBad(a + xml + b, 'Unable to parse unit kind: ')

        # Invalid unit attributes
        xml = (
            '<unitDefinition id="centimeter">'
            ' <listOfUnits>'
            '  <unit kind="volt" scale="made-up" />'
            ' </listOfUnits>'
            '</unitDefinition>')
        self.assertBad(a + xml + b, 'Unable to parse unit attributes')

    def test_tag(self):
        # Tests handling of different namespaces.

        # Create content that raises a missing ID error and therefore
        # tags the component.
        xml_content = (
            '<model>'
            ' <listOfCompartments>'
            '  <compartment />'
            ' </listOfCompartments>'
            '</model>')

        # Test SBML 3.2 namespace
        level = 3
        version = 2
        lv = 'level' + str(level) + '/version' + str(version)
        sbml_file = (
            '<sbml xmlns="http://www.sbml.org/sbml/' + lv + '/core"'
            ' level="' + str(level) + '"'
            ' version="' + str(version) + '">'
            + xml_content +
            '</sbml>')

        tag = 'sbml:compartment'
        message = 'Element ' + tag + ' is missing required'

        self.assertRaisesRegex(
            SBMLParsingError, message, self.p.parse_string, sbml_file)


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#
# Myokit symbolic expression classes. Defines different expressions, equations
# and the unit system.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import traceback

from collections import OrderedDict

import myokit
import myokit.formats.ansic
import myokit.pype

# Location of source file
SOURCE_FILE = 'cmodel.h'


class CModel:
    """
    Generates ansi-C code for a model.

    Example::

        model = myokit.parse_model('example')
        cmodel = CModel(model)
        print(cmodel.code)

    Arguments:

    ``model``
        A :class:`myokit.Model`.
    ``pacing_labels``
        A list of ``str`` variable labels, each corresponding to a paced
        variable.
    ``sensitivities``
        Either ``None`` or a tuple ``(dependents, independents)``. See
        :class:`myokit.Simulation` for details.

    The following properties are all public for easy access. But note that they
    do not interact with the compiled header so changing them will have little
    effect.

    ``model``
        The myokit.Model this CModel was compiled for.
    ``code``
        The generated model code. Can be placed in a header file or
        incorporated directly.

    Sensitivities:

    ``has_sensitivities``
        True if this model was created with sensitivity calculations enabled.
    ``dependents``
        A list of "dependent variables" for sensitivity calculations, all
        specified as expressions (:class:`myokit.Name` or
        :class:`myokit.Derivative`).
    ``independents``
        A list of "independent variables" for sensitivity calculations, all
        specified as expressions (:class:`myokit.Name` or
        :class:`myokit.InitialValue`).

    Constants (and parameters) are all stored inside ordered dicts mapping
    variable objects onto equations. Each dict is stored in a solvable order.

    ``parameters``
        Constants used as "independent variable" in sensitivity calculations.
    ``parameter_derived``
        Constants that depend on parameters (and possibly on literals too).
    ``literals``
        Constants that do not depend on any other constants (and are not
        parameters).
    ``literal_derived``
        Constants that depend on literals, but not on parameters.

    """
    def __init__(self, model, pacing_labels, sensitivities):

        # Parse sensitivity arguments
        has_sensitivities, dependents, independents = \
            self._parse_sensitivities(model, sensitivities)

        # Set unique names in model: variable names will be prepended so don't
        # bother with keywords.
        model.create_unique_names()

        # Get mapping from variables to C variable names as used in model.h
        labels = {
            'time': 'time',
            'realtime': 'realtime',
            'evaluations': 'evaluations',
        }
        for i, label in enumerate(pacing_labels):
            labels[label] = 'pace_values[' + str(i) + ']'
        bound_variables = myokit._prepare_bindings(model, labels)

        # Get equations in solvable order (grouped by component)
        equations = model.solvable_order()

        # Derive sensitivity equations
        if has_sensitivities:
            output_equations = self._derive_sensitivity_equations(
                model, dependents, independents)
        else:
            output_equations = []

        # Partition constants into 4 types

        literals, literal_derived, parameters, parameter_derived = \
            self._partition_constants(equations, independents)

        # Get naming function and expression writer
        v, w = self._create_expression_writer(independents, parameters)

        # Generate code
        code = self._generate_code(
            model, equations, bound_variables, dependents, independents,
            output_equations, literals, literal_derived, parameters,
            parameter_derived, v, w)

        # Provide public properties
        self.model = model
        self.code = code
        self.has_sensitivities = has_sensitivities
        self.dependents = dependents
        self.independents = independents

        # Literals, literal-derived, parameters, and parameter-derived, all in
        # solvable order. Parameters use the ordering given in `independents`
        # (which is sortable, as parameters are independent).
        self.literals = literals
        self.literal_derived = literal_derived
        self.parameters = parameters
        self.parameter_derived = parameter_derived

    def _parse_sensitivities(self, model, sensitivities):
        """
        Parses the ``sensitivities`` constructor argument and returns a tuple
        ``(has_sensitivities, dependents, independents)``, where
        ``has_sensitivities`` is a boolean and the other two entries are lists
        containing :class:`myokit.Expression` objects.

        Acceptable input for dependents (y in dy/dx):

        - Variable (state or intermediary)
        - Name or Derivative
        - "ina.INa" or "dot(membrane.V)"

        Acceptable input for independents (x in dy/dx):

        - Variable (literal)
        - Name or InitialValue
        - "ikr.gKr" or "init(membrane.V)"

        The resulting lists contain Expression objects.
        """
        if sensitivities is None:
            return False, [], []

        # Get lists
        try:
            deps, indeps = sensitivities
            deps = list(deps)
            indeps = list(indeps)
        except Exception:
            raise ValueError(
                'The argument `sensitivities` must be None, or a tuple'
                ' containing two lists.')
        if len(deps) == 0 or len(indeps) == 0:
            return False, [], []

        # Create output lists
        dependents = []
        independents = []

        # Check dependents, make sure all are Name or Derivative objects from
        # the cloned model.
        for x in deps:
            deriv = False
            if isinstance(x, myokit.Variable):
                var = model.get(x.qname())
            elif isinstance(x, myokit.Name):
                var = model.get(x.var().qname())
            elif isinstance(x, myokit.Derivative):
                deriv = True
                var = model.get(x.var().qname())
            else:
                x = str(x)
                if x[:4] == 'dot(' and x[-1:] == ')':
                    deriv = True
                    var = x = x[4:-1]
                var = model.get(x)
            lhs = myokit.Name(var)
            if deriv:
                lhs = myokit.Derivative(lhs)
                if not var.is_state():
                    raise ValueError(
                        'Sensitivity of ' + lhs.code() + ' requested, but '
                        + var.qname() + ' is not a state variable.')
            elif var.is_bound():
                raise ValueError(
                    'Sensitivities cannot be calculated for bound'
                    ' variables (got ' + str(var.qname()) + ').')
            # Note: constants are fine, just not very useful! But may be
            # easy, e.g. when working with multiple models.
            dependents.append(lhs)

        # Check independents, make sure all are Name or InitialValue
        # objects from the cloned model.
        for x in indeps:
            init = False
            if isinstance(x, myokit.Variable):
                var = model.get(x.qname())
            elif isinstance(x, myokit.Name):
                var = model.get(x.var().qname())
            elif isinstance(x, myokit.InitialValue):
                init = True
                var = model.get(x.var().qname())
            else:
                x = str(x)
                if x[:5] == 'init(' and x[-1:] == ')':
                    init = True
                    x = x[5:-1]
                var = model.get(x)
            lhs = myokit.Name(var)
            if init:
                lhs = myokit.InitialValue(myokit.Name(var))
                if not var.is_state():
                    raise ValueError(
                        'Sensitivity with respect to ' + lhs.code() +
                        ' requested, but ' + var.qname() + ' is not a'
                        ' state variable.')
            elif not var.is_literal():
                raise ValueError(
                    'Sensitivity with respect to ' + var.qname() +
                    ' requested, but this is not a literal variable (it'
                    ' depends on other variables).')
            independents.append(lhs)

        return True, dependents, independents

    def _derive_sensitivity_equations(self, model, deps, indeps):
        """
        Derive expressions needed to evaluate the variables we want to output
        partial derivatives of.
        """
        # Derive equations needed to calculate the requested sensitivities,
        # assuming that the sensitivities of the state variables are known.
        s_output_equations = []

        # First, get variables instead of LhsExpressions. Ignore Name(state),
        # as we already have these, and convert Derivative(Name(state)) into
        # variables.
        output_variables = [
            lhs.var() for lhs in deps
            if not (isinstance(lhs, myokit.Name) and lhs.var().is_state())]

        # Now call expressions_for, which will return expressions to evaluate
        # the rhs for each variable (i.e. the dot(x) rhs for states).
        output_equations, _ = model.expressions_for(*output_variables)
        del output_variables, _

        # Gather output expressions for each parameter or initial value we want
        # sensitivities w.r.t.
        for expr in indeps:
            eqs = []
            for eq in output_equations:
                rhs = eq.rhs.diff(expr, independent_states=False)
                if rhs.is_number(0) and eq.lhs not in deps:
                    continue
                lhs = myokit.PartialDerivative(eq.lhs, expr)
                eqs.append(myokit.Equation(lhs, rhs))
            s_output_equations.append(eqs)

        return s_output_equations

    def _partition_constants(self, equations, independents):
        """
        Partitions the model's constants into four (non-overlapping) groups,
        each stored in solvable order.
        """
        # NOTE: Some methods in cmodel.h require the ordering of parameters and
        #       literals to match that in the independents ordered-dict.
        literals = OrderedDict()
        literal_derived = OrderedDict()
        parameters = OrderedDict()
        parameter_derived = OrderedDict()

        # Get all parameters, in the same order as independents
        p = set()
        for lhs in independents:
            if isinstance(lhs, myokit.Name):
                p.add(lhs)
                parameters[lhs.var()] = None

        # Scan over equations
        for label, eqs in equations.items():
            for eq in eqs.equations(const=True):
                var = eq.lhs.var()
                if eq.lhs in p:
                    parameters[var] = eq
                elif var.is_literal():
                    literals[var] = eq
                elif eq.rhs.references().intersection(p):
                    p.add(eq.lhs)
                    parameter_derived[var] = eq
                else:
                    literal_derived[var] = eq

        return literals, literal_derived, parameters, parameter_derived

    def _create_expression_writer(self, independents, parameters):
        """
        Creates a variable/expression naming function and an expression writer.
        """

        def v(var):
            """
            Returns a readable variable name to use in code (which will be
            mapped to a model entry by a C define).

            C Variable names:

            - State: Y_x
            - Derivative: D_x
            - Bound: B_x
            - Intermediary: V_x
            - Parameter: P_x
            - Other constant: C_x

            Where `x` is the uname of the Myokit variable.

            For sensitivities `x` represents the name of the independent
            variable (x in dy/dx), and `i` represents the index of the
            dependent variable in the list of dependent variables. This is used
            in place of a name, because that would require stringing two unames
            together in some way that ensures the result does not overlap with
            any unames.

            - State sensitivity: Si_Y_x
            - Derivative sensitivity: Si_D_x (todo)
            - Intermediary sensitivity: Si_V_x

            """
            if not isinstance(var, (myokit.LhsExpression, myokit.Variable)):
                raise ValueError(  # pragma: no cover
                    'v() called with ' + str(var) + ' of type '
                    + str(type(var)))

            # Partial derivative
            if isinstance(var, myokit.PartialDerivative):
                i = var.dependent_expression()
                j = str(independents.index(var.independent_expression()))
                if isinstance(i, myokit.Derivative):
                    return 'S' + j + '_D_' + i.var().uname()
                if i.var().is_state():
                    return 'S' + j + '_Y_' + i.var().uname()
                return 'S' + j + '_V_' + i.var().uname()

            # Derivative
            if isinstance(var, myokit.Derivative):
                return 'D_' + var.var().uname()

            # Name given? get variable object from name
            if isinstance(var, myokit.Name):
                var = var.var()

            # State
            if var.is_state():
                return 'Y_' + var.uname()

            # Parameter
            if var in parameters:
                return 'P_' + var.uname()

            # Constant (parameter derived, literal, literal derived)
            if var.is_constant():
                return 'C_' + var.uname()

            # Bound variable
            if var.is_bound():
                return 'B_' + var.binding()

            # Intermediary variable
            return 'V_' + var.uname()

        # Create expression writer
        w = myokit.formats.ansic.AnsiCExpressionWriter()
        w.set_lhs_function(v)

        return v, w

    def _generate_code(
            self, model, equations, bound_variables, dependents, independents,
            output_equations, literals, literal_derived, parameters,
            parameter_derived, v, w):
        """ Generates and returns the model code. """

        # Get states whose initial value is used in sensivitity calculations
        initials = [p.var().index() for p in independents
                    if isinstance(p, myokit.InitialValue)]

        # Arguments
        args = {
            'model': model,
            'equations': equations,
            'bound_variables': bound_variables,
            's_dependents': dependents,
            's_independents': independents,
            's_output_equations': output_equations,
            'initials': initials,
            'parameters': parameters,
            'parameter_derived': parameter_derived,
            'literals': literals,
            'literal_derived': literal_derived,
            'v': v,
            'w': w,
        }

        # Path
        path = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Parse template
        p = myokit.pype.TemplateEngine()
        try:
            return p.process(path, args)
        except myokit.pype.PypeError:  # pragma: no cover
            msg = ['An error ocurred while processing the template']
            msg.append(traceback.format_exc())
            extra = p.error_details()
            if extra:
                msg.append(extra)
            raise myokit.GenerationError('\n'.join(msg))




#
# Python expression writer and keywords
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class PythonExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their Python
    equivalent.
    """
    def __init__(self):
        super().__init__()

        self._flhs = None
        self.set_lhs_function(lambda v: str(v))

        self._function_prefix = 'math.'

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         ``myokit.LhsExpression`` object.

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        self._flhs = f

    def _ex_infix(self, e, op):
        """
        Handles ex() for infix operators
        """
        if e.bracket(e[0]):
            out = '(' + self.ex(e[0]) + ') ' + op
        else:
            out = self.ex(e[0]) + ' ' + op
        if e.bracket(e[1]):
            return out + ' (' + self.ex(e[1]) + ')'
        else:
            return out + ' ' + self.ex(e[1])

    def _ex_function(self, e, func):
        """
        Handles ex() for function operators
        """
        return self._function_prefix + func \
            + '(' + ', '.join([self.ex(x) for x in e]) + ')'

    def _ex_infix_condition(self, e, op):
        """
        Handles ex() for infix condition operators
        """
        return '(' + self.ex(e[0]) + ' ' + op + ' ' + self.ex(e[1]) + ')'

    def _ex_name(self, e):
        return self._flhs(e)

    def _ex_derivative(self, e):
        return self._flhs(e)

    def _ex_initial_value(self, e):
        return self._flhs(e)

    def _ex_partial_derivative(self, e):
        return self._flhs(e)

    def _ex_number(self, e):
        return myokit.float.str(e)

    def _ex_prefix_plus(self, e):
        return self.ex(e[0])

    def _ex_prefix_minus(self, e):
        if e.bracket(e[0]):
            return '(-(' + self.ex(e[0]) + '))'
        else:
            return '(-' + self.ex(e[0]) + ')'

    def _ex_plus(self, e):
        return self._ex_infix(e, '+')

    def _ex_minus(self, e):
        return self._ex_infix(e, '-')

    def _ex_multiply(self, e):
        return self._ex_infix(e, '*')

    def _ex_divide(self, e):
        return self._ex_infix(e, '/')

    def _ex_quotient(self, e):
        return self._ex_infix(e, '//')

    def _ex_remainder(self, e):
        return self._ex_infix(e, '%')

    def _ex_power(self, e):
        return self._ex_infix(e, '**')

    def _ex_sqrt(self, e):
        return self._ex_function(e, 'sqrt')

    def _ex_sin(self, e):
        return self._ex_function(e, 'sin')

    def _ex_cos(self, e):
        return self._ex_function(e, 'cos')

    def _ex_tan(self, e):
        return self._ex_function(e, 'tan')

    def _ex_asin(self, e):
        return self._ex_function(e, 'asin')

    def _ex_acos(self, e):
        return self._ex_function(e, 'acos')

    def _ex_atan(self, e):
        return self._ex_function(e, 'atan')

    def _ex_exp(self, e):
        return self._ex_function(e, 'exp')

    def _ex_log(self, e):
        return self._ex_function(e, 'log')

    def _ex_log10(self, e):
        return self._ex_function(e, 'log10')

    def _ex_floor(self, e):
        return self._ex_function(e, 'floor')

    def _ex_ceil(self, e):
        return self._ex_function(e, 'ceil')

    def _ex_abs(self, e):
        return 'abs(' + self.ex(e[0]) + ')'

    def _ex_not(self, e):
        return 'not (' + self.ex(e[0]) + ')'

    def _ex_equal(self, e):
        return self._ex_infix_condition(e, '==')

    def _ex_not_equal(self, e):
        return self._ex_infix_condition(e, '!=')

    def _ex_more(self, e):
        return self._ex_infix_condition(e, '>')

    def _ex_less(self, e):
        return self._ex_infix_condition(e, '<')

    def _ex_more_equal(self, e):
        return self._ex_infix_condition(e, '>=')

    def _ex_less_equal(self, e):
        return self._ex_infix_condition(e, '<=')

    def _ex_and(self, e):
        return self._ex_infix_condition(e, 'and')

    def _ex_or(self, e):
        return self._ex_infix_condition(e, 'or')

    def _ex_if(self, e):
        return '(' + self.ex(e._t) + ' if ' + self.ex(e._i) + ' else ' \
                   + self.ex(e._e) + ')'

    def _ex_piecewise(self, e):
        s = ''
        n = len(e) // 2
        for i in range(0, n):
            s += '(' + self.ex(e._e[i]) + ' if ' + self.ex(e._i[i]) + ' else '
        s += self.ex(e._e[n])
        s += ')' * n
        return s


class NumPyExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to Python expressions
    intended for use in NumPy arrays.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = 'numpy.'
    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):
    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):
    #def _ex_power(self, e):
    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):

    def _ex_asin(self, e):
        return self._ex_function(e, 'arcsin')

    def _ex_acos(self, e):
        return self._ex_function(e, 'arccos')

    def _ex_atan(self, e):
        return self._ex_function(e, 'arctan')
    #def _ex_exp(self, e):
    #def _ex_log(self, e):
    #def _ex_log10(self, e):
    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):
    #def _ex_abs(self, e):
    #def _ex_not(self, e):
    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):
    #def _ex_and(self, e):
    #def _ex_or(self, e):

    def _ex_if(self, e):
        return self._function_prefix + 'select([' + self.ex(e._i) + '], [' \
            + self.ex(e._t) + '], ' + self.ex(e._e) + ')'

    def _ex_piecewise(self, e):
        n = len(e._i)
        s = [self._function_prefix, 'select([']
        s.append(', '.join([self.ex(x) for x in e._i]))
        s.append('], [')
        s.append(', '.join([self.ex(x) for x in e._e[:-1]]))
        s.append('], ')
        s.append(self.ex(e._e[n]))
        s.append(')')
        return ''.join(s)




#
# Tools for running experiments on multiple models.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit


#
# Deprecated since 2020-06-14
#
import warnings
warnings.warn(
    'The module myokit.lib.multi is deprecated: it will be removed in future'
    ' versions of Myokit.'
)


def iterdir(path, guarantee_model_names=True):
    """
    Iterates over a directory yielding tuples ``(model, protocol)`` where
    ``model`` is a :class:`myokit.Model` and ``protocol`` is a
    :class:`myokit.Protocol`.

    Depending on the contents of the found files, some entries might not have a
    protocol. Files without a model are skipped.

    The results will be yielded ordered by filename.
    The method does not descend into child directories.

    If ``guarantee_model_names`` is ``True`` (default), models that do not
    specify a ``name`` meta-attribute will be given their filename as name.
    This ensures every model read by this method has a name meta-property.
    """
    # Fix path
    path = os.path.expanduser(os.path.abspath(path))

    if not os.path.isdir(path):
        raise ValueError('Given path is not a directory.')

    # Scan files
    for fname in sorted(os.listdir(path)):
        fpath = os.path.join(path, fname)

        # Check if it's a model file
        if not os.path.isfile(fpath):
            continue
        base, ext = os.path.splitext(fname)
        if ext != '.mmt':
            continue

        # Read model & protocol
        model, protocol, x = myokit.load(fpath)

        # Skip files without model
        if model is None:
            continue

        # Set name attribute
        if guarantee_model_names:
            if not model.name():
                model.meta['name'] = base

        # Yield
        yield model, protocol


def scandir(path):
    """
    Scans a directory using :meth:`iterdir` and returns a list of
    :class:`myokit.Model` objects and a list of :class:`myokit.Protocol`
    objects. The models (and corresponding protocols) will be ordered by model
    name.
    """
    names = []
    ms = {}
    ps = {}
    for model, protocol in iterdir(path):
        name = model.name()
        names.append(name)
        ms[name] = model
        ps[name] = protocol
    models = []
    protocols = []
    for name in sorted(names):
        models.append(ms[name])
        protocols.append(ps[name])
    return models, protocols


def time(model):
    """
    Deprecated alias of :meth:`Model.timex()`.
    """
    # Deprecated since 2019-09-09
    import warnings
    warnings.warn(
        'The method `myokit.lib.multi.time()` is deprecated and will be'
        ' removed in future versions of Myokit. Please use `Model.timex()`'
        ' instead.')
    return model.timex()


def label(model, label):
    """
    Deprecated alias of :meth:`Model.labelx()`.
    """
    # Deprecated since 2019-09-09
    import warnings
    warnings.warn(
        'The method `myokit.lib.multi.label()` is deprecated and will be'
        ' removed in future versions of Myokit. Please use `Model.labelx()`'
        ' instead.')
    return model.labelx(label)


def binding(model, binding):
    """
    Deprecated alias of :meth:`Model.bindingx()`.
    """
    # Deprecated since 2019-09-09
    import warnings
    warnings.warn(
        'The method `myokit.lib.multi.binding()` is deprecated and will be'
        ' removed in future versions of Myokit. Please use `Model.bindingx()`'
        ' instead.')
    return model.bindingx(binding)


def unit(variable, unit):
    """
    Checks if the given variable's unit can be converted into units `unit` and,
    if so, returns the appropriate conversion factor. If not, a
    :class:`myokit.IncompatibleModelError` is raised.

    Example::

        >>> import myokit
        >>> import myokit.lib.multi as multi
        >>> m,p,x = myokit.load('example')
        >>> print(multi.unit(m.label('membrane_potential'), myokit.units.V))
        0.001

    (Because a millivolt can be converted to a volt by multiplying by 0.001)

    """
    try:
        return myokit.Unit.convert(1, variable.unit(), unit)
    except myokit.IncompatibleUnitError:
        raise myokit.IncompatibleModelError(
            variable.model().name(),
            'Incompatible units: ' + str(variable.unit()) + ' and '
            + str(unit) + '.')




#
# Tools for working with Hodgkin-Huxley style ion channel models
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import numpy as np

import myokit


class HHModel:
    """
    Represents a Hodgkin-Huxley (HH)-style model of an ion channel, extracted
    from a :class:`myokit.Model`.

    The class assumes the HH model can be (re)written as::

        I = prod(x[i]**n[i]) * f(V, p)
        dot(x[i]) = (x_inf[i](V, p) - x[i]) / tau_x[i](V, p)

    for any number of states ``x[i]`` with steady-state ``x_inf[i]`` and time
    constant ``tau_x[i]``, where ``V`` is the membrane potential, ``p`` is a
    set of parameters, ``f`` is an arbitrary function (for example
    ``g_max * (V - E_rev)``).

    The model variables to treat as parameter are specified by the user when
    the model is created. Any other variables, for example state variables such
    as intercellular calcium or constants such as temperature, are fixed when
    the current model is created and can no longer be changed. Initial values
    written as expressions are evaluated when the model is made.

    The current variable is optional.

    Arguments:

    ``model``
        The :class:`myokit.Model` to extract a current model from.
    ``states``
        An ordered list of state variables (or state variable names) from
        ``model``. All remaining state variables will be frozen in place.
    ``parameters=None``
        A list of parameters to maintain in their symbolic form.
    ``current=None``
        The current model's current variable, or ``None``.
    ``vm=None``
        The variable indicating membrane potential. If set to ``None``
        (default) the method will search for a variable with the label
        ``membrane_potential``.

    Example::

        import myokit
        import myokit.lib.hh as hh

        # Load a model from disk
        model = myokit.load_model('some-model.mmt')

        # Select the relevant states and parameters
        states = [
            'ina.C1',
            'ina.C2',
            'ina.O',
            ...
            ]
        parameters = [
            'ina.p1',
            'ina.p2',
            ...
            ]
        current = 'ina.INa'

        # Extract an ion current model
        mm = hh.HHModel(model, states, parameters, current)

    Alternatively, a HHModel can be constructed based on a single component
    using the method :meth:`from_component() <HHModel.from_component()>`::

        import myokit
        import myokit.lib.hh as hh

        # Load a model from disk
        model = myokit.load_model('some-model.mmt')

        # Extract a Hodgkin-Huxley style channel model
        mm = hh.HHModel.from_component(model.get('ina'))

    """
    # NOTE: A HHModel must be immutable!
    # This ensures that the simulations don't have to clone it (which would be
    # costly and difficult).
    # A HHModel can return a function to calculate state values, but it never
    # updates its internal state in any way.
    def __init__(self, model, states, parameters=None, current=None, vm=None):
        super().__init__()

        #
        # Check input
        #
        if not isinstance(model, myokit.Model):
            raise ValueError('First argument must be a myokit.Model.')

        # Check membrane potential variable is known, and is a qname
        if vm is None:
            vm = model.label('membrane_potential')
        if vm is None:
            raise HHModelError(
                'A membrane potential must be specified as `vm` or using the'
                ' label `membrane_potential`.')
        if isinstance(vm, myokit.Variable):
            vm = vm.qname()

        # Ensure all HH-states in the model are written in inf-tau form
        # This returns a clone of the original model
        self._model = convert_hh_states_to_inf_tau_form(model, vm)
        del model

        # Check and collect state variables
        self._states = []
        for state in states:
            if isinstance(state, myokit.Variable):
                state = state.qname()
            try:
                state = self._model.get(str(state), myokit.Variable)
            except KeyError:
                raise HHModelError('Unknown state: <' + str(state) + '>.')
            if not state.is_state():
                raise HHModelError(
                    'Variable <' + state.qname() + '> is not a state.')
            if state in self._states:
                raise HHModelError(
                    'State <' + state.qname() + '> was added twice.')
            self._states.append(state)
        del states

        # Check and collect parameter variables
        unique = set()
        self._parameters = []
        if parameters is None:
            parameters = []
        for parameter in parameters:
            if isinstance(parameter, myokit.Variable):
                parameter = parameter.qname()
            try:
                parameter = self._model.get(parameter, myokit.Variable)
            except KeyError:
                raise HHModelError(
                    'Unknown parameter: <' + str(parameter) + '>.')
            if not parameter.is_literal():
                raise HHModelError(
                    'Unsuitable parameter: <' + str(parameter) + '>.')
            if parameter in unique:
                raise HHModelError(
                    'Parameter listed twice: <' + str(parameter) + '>.')
            unique.add(parameter)
            self._parameters.append(parameter)
        del unique
        del parameters

        # Check current variable
        if current is not None:
            if isinstance(current, myokit.Variable):
                current = current.qname()
            current = self._model.get(current, myokit.Variable)
            if current.is_state():
                raise HHModelError('Current variable can not be a state.')
        self._current = current
        del current

        # Check membrane potential variable
        self._membrane_potential = self._model.get(vm)
        if self._membrane_potential in self._parameters:
            raise HHModelError(
                'Membrane potential should not be included in the list of'
                ' parameters.')
        if self._membrane_potential in self._states:
            raise HHModelError(
                'The membrane potential should not be included in the list of'
                ' states.')
        if self._membrane_potential == self._current:
            raise HHModelError(
                'The membrane potential should not be the current variable.')
        del vm

        # Get values of all states
        # Note: Do this _before_ changing the model!
        s = self._model.initial_values(True)
        self._model.set_initial_values(s)  # Remove expressions
        self._default_state = np.array(
            [v.initial_value(True) for v in self._states])

        #
        # Demote unnecessary states and remove bindings
        #

        # Freeze remaining, non-current-model states
        # Note: list() cast is required so that we iterate over a static list,
        # otherwise we can get issues because the iterator depends on the model
        # (which we're changing).
        for k, state in enumerate(list(self._model.states())):
            if state not in self._states:
                state.demote()
                state.set_rhs(s[k])

        # Unbind everything except time
        for label, var in self._model.bindings():
            if label != 'time':
                var.set_binding(None)

        # Check if current variable depends on selected states
        # (At this point, anything not dependent on the states is a constant)
        if self._current is not None and self._current.is_constant():
            raise HHModelError(
                'Current must be a function of the selected state variables.')

        # Ensure all states are written in inf-tau form
        for state in self._states:
            if not has_inf_tau_form(state, self._membrane_potential):
                raise HHModelError(
                    'State `' + state.qname() + '` must have "inf-tau form" or'
                    ' "alpha-beta form". See'
                    ' `myokit.lib.hh.has_inf_tau_form()` and'
                    ' `myokit.lib.hh.has_alpha_beta_form()`.'
                )

        #
        # Remove unused variables from internal model, and evaluate any
        # literals.
        #
        # 1. Make sure that current variable is maintained by temporarily
        #    setting it as a state variable.
        # 2. Similarly, make sure parameters and membrane potential are not
        #    evaluated and/or removed
        #
        self._membrane_potential.promote(0)
        if self._current is not None:
            self._current.promote(0)
        for p in self._parameters:
            p.promote(0)

        # Evaluate all constants and remove unused variables
        for var in self._model.variables(deep=True, const=True):
            var.set_rhs(var.rhs().eval())
        self._model.validate(remove_unused_variables=True)

        self._membrane_potential.demote()
        for p in self._parameters:
            p.demote()
        if self._current is not None:
            self._current.demote()

        # Validate modified model
        self._model.validate()

        #
        # Create functions
        #

        # Create a list of inputs to the functions
        self._inputs = [self._membrane_potential] + self._parameters

        # Get the default values for all inputs
        self._default_inputs = np.array([v.eval() for v in self._inputs])

        #
        # Create a function that calculates the states analytically
        #
        # The created self._function has signature _f(_y0, _t, v, params*)
        # and returns a tuple (states, current). If _t is a scalar, states is a
        # sequence of state values and current is a single current value. If _t
        # is a numpy array of times, states is a sequence of arrays, and
        # current is a numpy array. If no current variable is known current is
        # always None.
        #
        f = []
        args = ['_y0', '_t'] + [i.uname() for i in self._inputs]
        f.append('def _f(' + ', '.join(args) + '):')
        f.append('_y = [0]*' + str(len(self._states)))

        # Add equations to calculate all infs and taus
        w = myokit.numpy_writer()
        order = self._model.solvable_order()
        ignore = set(self._inputs + self._states + [self._model.time()])
        for group in order.values():
            for eq in group:
                var = eq.lhs.var()
                if var in ignore or var == self._current:
                    continue
                f.append(w.eq(eq))

        # Add equations to calculate updated state
        for k, state in enumerate(self._states):
            inf, tau = get_inf_and_tau(state, self._membrane_potential)
            inf = w.ex(myokit.Name(inf))
            tau = w.ex(myokit.Name(tau))
            k = str(k)
            f.append(
                '_y[' + k + '] = ' + state.uname() + ' = '
                + inf + ' + (_y0[' + k + '] - ' + inf
                + ') * numpy.exp(-_t / ' + tau + ')')

        # Add current calculation
        if self._current is not None:
            f.append('_i = ' + w.ex(self._current.rhs()))
        else:
            f.append('_i = None')

        # Add return statement and create python function from f
        f.append('return _y, _i')
        for i in range(1, len(f)):
            f[i] = '    ' + f[i]
        f = '\n'.join(f)
        #print(f)
        local = {}
        exec(f, {'numpy': np}, local)
        self._function = local['_f']

        #
        #
        # Create a function that calculates the steady states
        #
        #
        g = []
        args = [i.uname() for i in self._inputs]
        g.append('def _g(' + ', '.join(args) + '):')
        g.append('_y = [0]*' + str(len(self._states)))
        for k, state in enumerate(self._states):
            k = str(k)
            inf, tau = get_inf_and_tau(state, self._membrane_potential)
            inf = inf.rhs().clone(expand=True, retain=self._inputs)
            g.append('_y[' + str(k) + '] = ' + w.ex(inf))

        # Create python function from g
        g.append('return _y')
        for i in range(1, len(g)):
            g[i] = '    ' + g[i]
        g = '\n'.join(g)
        local = {}
        exec(g, {'numpy': np}, local)
        self._steady_state_function = local['_g']

    def current(self):
        """
        Returns the name of the current variable used by this model, or None if
        no current variable was specified.
        """
        return self._current

    def default_membrane_potential(self):
        """
        Returns this HH model's default membrane potential value.
        """
        return self._default_inputs[0]

    def default_parameters(self):
        """
        Returns this HH model's default parameter values
        """
        return list(self._default_inputs[1:])

    def default_state(self):
        """
        Returns this HH model's default state values.
        """
        return list(self._default_state)

    @staticmethod
    def from_component(
            component, states=None, parameters=None, current=None, vm=None):
        """
        Creates a current model from a component, using the following rules:

          1. Every state in the component is a state in the current model.
          2. Every (non-nested) constant in the component is a parameter.
          3. The component should contain exactly one non-nested intermediary
             variable whose value depends on the model states, this will be
             used as the current variable.
          4. The model contains a variable labeled "membrane_potential".

        Any of the automatically set variables can be overridden using the
        keyword arguments ``states``, ``parameters``, ``current`` and ``vm``.

        The parameters, if determined automatically, will be specified in
        alphabetical order (using a natural sort).
        """
        model = component.model()

        # Get or check states
        if states is None:

            # Get state variables
            states = [x for x in component.variables(state=True)]

            # Sort by state index
            states.sort(key=lambda x: x.index())

        else:

            # Make sure states are variables. This is required to automatically
            # find a current variable.
            states_in = states
            states = []
            for state in states_in:
                if isinstance(state, myokit.Variable):
                    state = state.qname()
                states.append(model.get(state))

        # Get parameters
        if parameters is None:

            # Get parameters
            parameters = [
                x for x in component.variables(const=True) if x.is_literal()]
            # Sort by qname, using natural sort
            parameters.sort(
                key=lambda x: myokit.tools.natural_sort_key(x.qname()))

        # Get current
        if current is None:
            currents = []
            for x in component.variables(inter=True):
                for y in x.refs_to(state_refs=True):
                    if y in states:
                        # Found a candidate!
                        currents.append(x)
                        break
            if len(currents) > 1:
                raise HHModelError(
                    'The given component has more than one variable that could'
                    ' be a current: '
                    + ', '.join(['<' + x.qname() + '>' for x in currents])
                    + '.')
            try:
                current = currents[0]
            except IndexError:
                pass

        # Get membrane potential
        if vm is None:
            vm = model.label('membrane_potential')
            if vm is None:
                raise HHModelError(
                    'The model must define a variable labeled as'
                    ' "membrane_potential".')

        # Create and return LinearModel
        return HHModel(model, states, parameters, current, vm)

    def function(self):
        """
        Returns a function that evaluates the state values at any time.

        The returned function has the signature::

            f(y0, t, p1, p2, p3, ..., V) --> (states, current)

        where ``y0`` is a sequence containing the state values at time 0, where
        ``t`` is a scalar or numpy array of times at which to evaluate, where
        ``p1, p2, p3, ...`` are the model parameters, and where ``V`` is the
        membrane potential.

        The function output is always a tuple ``(states, current)``. If ``t``
        is a single time, then ``states`` will be a list of values (one per
        model state) and current will be a scalar (or ``None`` if no current
        variable was specified). If ``t`` is a numpy array of times then
        ``states`` will be a list of numpy arrays, and current will be a numpy
        array (or ``None`` if no current variable was specified).
        """
        return self._function

    def membrane_potential(self):
        """
        Returns the name of the membrane potential variable used by this model.
        """
        return self._membrane_potential.qname()

    def parameters(self):
        """
        Returns the names of the parameter variables used by this model.
        """
        return [v.qname() for v in self._parameters]

    def _parse_inputs(self, membrane_potential, parameters):
        """
        Returns a list of inputs based on the given V and parameters. If either
        is ``None`` it will be replaced by the default values.
        """
        inputs = list(self._default_inputs)
        if membrane_potential is not None:
            inputs[0] = float(membrane_potential)
        if parameters is not None:
            if len(parameters) != len(self._parameters):
                raise ValueError(
                    'Illegal parameter vector size: '
                    + str(len(self._parameters)) + ' required, '
                    + str(len(parameters)) + ' provided.')
            inputs[1:] = [float(x) for x in parameters]
        return inputs

    def reduced_model(self):
        """
        Returns a reduced :class:`myokit.Model`, containing only the parts
        necessary to calculate the specified states and current.
        """
        return self._model.clone()

    def states(self):
        """
        Returns the names of the state variables used by this model.
        """
        return [v.qname() for v in self._states]

    def steady_state(self, membrane_potential=None, parameters=None):
        """
        Returns the steady state solution for this model.

        Arguments:

        ``membrane_potential``
            The value to use for the membrane potential, or ``None`` to use the
            value from the original :class:`myokit.Model`.
        ``parameters``
            The values to use for the parameters, given in the order they were
            originally specified in (if the model was created using
            :meth:`from_component()`, this will be alphabetical order), or
            ``None`` to use the values from the original model.

        Returns a list of steady state values.
        """
        inputs = self._parse_inputs(membrane_potential, parameters)
        return self._steady_state_function(*inputs)


class AnalyticalSimulation:
    """
    Analytically evaluates a :class:`HHModel`'s state for a given set of points
    in time.

    Each simulation object maintains an internal state consisting of

    * The current simulation time
    * The current state
    * The default state

    When a simulation is created, the simulation time is set to zero and both
    the current and default state are initialized using the :class:`HHModel`.
    After each call to :meth:`run()` the time and current state are updated,
    so that each successive call to run continues where the previous simulation
    left off.

    A :class:`protocol <myokit.Protocol>` can be used to set the membrane
    potential during the simulation, or the membrane potential can be adjusted
    manually between runs.

    Example::

        import myokit
        import myokit.lib.hh as hh

        # Create an ion current model
        m = myokit.load_model('luo-1991.mmt')
        m = hh.HHModel.from_component(m.get('ina'))

        # Create an analytical simulation object
        s = hh.AnalyticalSimulation(m)

        # Run a simulation
        s.set_membrane_potential(-30)
        d = s.run(10)

        # Show the results
        import matplotlib.pyplot as plt
        plt.figure()
        plt.subplot(211)
        for state in m.states():
            plt.plot(d.time(), d[state], label=state)
        plt.legend(loc='center right')
        plt.subplot(212)
        plt.plot(d.time(), d[m.current()])
        plt.show()

    """
    def __init__(self, model, protocol=None):
        super().__init__()

        # Check model
        if not isinstance(model, HHModel):
            raise ValueError('First parameter must be a `HHModel`.')
        self._model = model

        # Check protocol
        if protocol is None:
            self._protocol = None
        elif not isinstance(protocol, myokit.Protocol):
            raise ValueError('Protocol must be a myokit.Protocol object')
        else:
            self._protocol = protocol.clone()

        # Time variable
        self._time = 0

        # Check if we have a current variable
        self._has_current = self._model.current() is not None

        # Set state
        self._state = np.array(
            self._model.default_state(), copy=True, dtype=float)

        # Set default state
        self._default_state = np.array(self._state, copy=True)

        # Get membrane potential
        self._membrane_potential = self._model.default_membrane_potential()

        # Get parameters
        self._parameters = np.array(
            self._model.default_parameters(), copy=True, dtype=float)

        # Mapping from parameter names to index in parameter array
        self._parameter_map = {}
        for i, p in enumerate(self._model.parameters()):
            self._parameter_map[p] = i

        # Get function
        self._function = self._model.function()

        # If protocol was given, create pacing system, update vm
        self._pacing = None
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

        # Keys for logging
        self._time_key = self._model._model.time().qname()
        self._vm_key = self._model._membrane_potential
        self._log_keys = [self._time_key, self._vm_key] + self._model.states()
        if self._has_current:
            self._log_keys.append(self._model.current())

    def default_state(self):
        """
        Returns the default state used by this simulation.
        """
        return list(self._default_state)

    def membrane_potential(self):
        """
        Returns the currently set membrane potential.
        """
        return self._membrane_potential

    def parameters(self):
        """
        Returns the currently set parameter values.
        """
        return list(self._parameters)

    def pre(self, duration):
        """
        Performs an unlogged simulation for ``duration`` time units and uses
        the final state as the new default state.

        After the simulation:

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will set the current
        state to this new default state.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')

        # Run
        # This isn't much faster, but this way the simulation's interface is
        # similar to the standard simulation one.
        old_time = self._time
        self.run(duration, log_interval=2 * duration)

        # Update default state
        self._default_state = np.array(self._state, copy=True)

        # Reset time, reset protocol
        self._time = old_time
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to zero.
        - The state is set to the default state.

        """
        self._time = 0
        self._state = np.array(self._default_state, copy=True)
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def run(self, duration, log=None, log_interval=0.01, log_times=None):
        """
        Runs a simulation for ``duration`` time units.

        After the simulation:

        - The simulation time will be increased by ``duration`` time units.
        - The simulation state will be updated to the last reached state.

        Arguments:

        ``duration``
            The number of time units to simulate.
        ``log``
            A log from a previous run can be passed in, in which case the
            results will be appended to this one.
        ``log_interval``
            The time between logged points.
        ``log_times``
            A pre-defined sequence of times to log at. If set, ``log_interval``
            will be ignored.

        Returns a :class:`myokit.DataLog` with the simulation results.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')
        log_interval = float(log_interval)
        if log_interval <= 0:
            raise ValueError('Log interval must be greater than zero.')

        # Check log_times
        if log_times is None:
            log_times = self._time + np.arange(0, duration, log_interval)

        # Set up logging
        if log is None:

            # Create new log
            log = myokit.DataLog()
            log.set_time_key(self._time_key)
            for key in self._log_keys:
                log[key] = np.zeros(log_times.shape)
            offset = 0

        else:

            # Check existing log
            if len(log.keys()) > len(self._log_keys):
                raise ValueError('Invalid log: contains extra keys.')
            try:
                key = self._time_key    # Note: error msg below uses `key`
                offset = len(log[key])
                for key in self._log_keys:
                    log[key] = np.concatenate((
                        log[key], np.zeros(log_times.shape)))
            except KeyError:
                raise ValueError(
                    'Invalid log: missing entry for <' + str(key) + '>.')

        # Run simulation
        if self._protocol is None:

            # User defined membrane potential
            self._run(log, log_times, self._time + duration, offset)

        else:

            # Voltage clamp
            tfinal = self._time + duration
            while self._time < tfinal:
                # Run simulation
                tnext = min(tfinal, self._pacing.next_time())
                times = log_times[np.logical_and(
                    log_times >= self._time, log_times < tnext)]
                self._run(log, times, tnext, offset)
                offset += len(times)

                # Update pacing
                self._membrane_potential = self._pacing.advance(tnext)

        # Return
        return log

    def _run(self, log, times, tnext, offset):
        """
        Runs a simulation with the current membrane potential.

        Arguments:

        ``log``
            The log to append to.
        ``times``
            The times to evaluate at.
        ``tnext``
            The final time to move to.
        ``offset``
            The offset in the ``times`` array to log in

        """
        # Simulate with fixed V
        if self._has_current:
            states, currents = self.solve(times - self._time)
        else:
            states = self.solve(times - self._time)

        # Log results
        lo = offset
        hi = lo + len(times)
        key = log.time_key()
        log[key][lo:hi] = times
        for i, key in enumerate(self._model.states()):
            log[key][lo:hi] = states[i]
        if self._has_current:
            key = self._model.current()
            log[key][lo:hi] = currents
        key = self._model._membrane_potential
        log[key][lo:hi] = self._membrane_potential

        # Now run simulation for final time (which might not be included in the
        # list of logged times, and should not, if you want to be able to
        # append logs without creating duplicate points).
        times = np.array([tnext - self._time])
        if self._has_current:
            states, currents = self.solve(times)
        else:
            states = self.solve(times)

        # Update simulation state
        self._state = np.array(states[:, -1], copy=True)
        self._time = tnext

    def set_constant(self, variable, value):
        """
        Updates a single parameter to a new value.
        """
        self._parameters[self._parameter_map[variable]] = float(value)

    def set_default_state(self, state):
        """
        Changes this simulation's default state.
        """
        state = np.array(state, copy=True, dtype=float)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.any(state < 0) or np.any(state > 1):
            raise ValueError(
                'All states must be in the range [0, 1].')
        self._default_state = state

    def set_membrane_potential(self, v):
        """
        Changes the membrane potential used in this simulation.
        """
        if self._protocol:
            raise Exception(
                'Membrane potential cannot be set if a protocol is used.')
        self._membrane_potential = float(v)

    def set_parameters(self, parameters):
        """
        Changes the parameter values used in this simulation.
        """
        if len(parameters) != len(self._parameters):
            raise ValueError(
                'Wrong size parameter vector, expecting ('
                + str(len(self._parameters)) + ') values.')
        self._parameters = np.array(parameters, copy=True, dtype=float)

    def set_state(self, state):
        """
        Changes the initial state used by in this simulation.
        """
        state = np.array(state, copy=True, dtype=float)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.any(state < 0) or np.any(state > 1):
            raise ValueError(
                'All states must be in the range [0, 1].')
        self._state = state

    def solve(self, times):
        """
        Evaluates and returns the states at the given times.

        In contrast to :meth:`run()`, this method simply evaluates the states
        (and current) at the given times, using the last known settings for
        the state and membrane potential. It does not use a protocol and does
        not take into account the simulation time. After running this method,
        the state and simulation time are *not* updated.

        Arguments:

        ``times``
            A series of times, where each time must be some ``t >= 0``.

        For models with a current variable, this method returns a tuple
        ``(state, current)`` where ``state`` is a matrix of shape
        ``(len(states), len(times))`` and ``current`` is a vector
        of length ``len(times)``.

        For models without a current variable, only ``state`` is returned.
        """
        states, current = self._function(
            self._state,
            times,
            self._membrane_potential,
            *self._parameters
        )

        # Calculate states or states + current
        states = np.array(states)
        if self._has_current:
            return states, current
        else:
            return states

    def state(self):
        """
        Returns the initial state used by this simulation.
        """
        return list(self._state)


def convert_hh_states_to_inf_tau_form(model, v=None):
    """
    Scans a :class:`myokit.Model` for Hodgkin-Huxley style states written in
    "alpha-beta form", and converts them to "inf-tau form".

    For any state ``x`` written in the form
    ``dot(x) = alpha * (1 - x) - beta * x`` this method will calculate the
    steady state and time constant, and add variables for both. Next, the state
    RHS will be replaced by an expression of the form
    ``dot(x) = (x_inf - x) / tau_x``, where ``x_inf`` and ``tau_x`` are the
    new variables.

    See also: :meth:`get_alpha_and_beta()`.

    Arguments:

    ``model``
        A :class:`myokit.Model` object to convert.
    ``v``
        An optional :class:`myokit.Variable`` representing the membrane
        potential. If not given, the method will search for a variable labelled
        ``membrane_potential``. An error is raised if no membrane potential
        variable can be found.

    Returns an updated copy of the given model.
    """
    # Clone the model before editing
    if not isinstance(model, myokit.Model):
        raise ValueError('Given `model` must be a myokit.Model.')
    model = model.clone()

    # Check membrane potential variable is known.
    if v is None:
        v = model.label('membrane_potential')
        if v is None:
            raise ValueError(
                'Membrane potential must be given as `v` or by setting the'
                ' label `membrane_potential` in the model.')
    else:
        # Ensure v is a variable, and from the cloned model
        if isinstance(v, myokit.Variable):
            v = v.qname()
        v = model.get(v)

    # Loop over states
    # - If they're in alpha-beta form, add new (nested) variables inf and tau
    for x in model.states():
        res = get_alpha_and_beta(x, v)
        if res is not None:
            # Create variabless for inf and tau
            a = myokit.Name(res[0])
            b = myokit.Name(res[1])
            tau = x.add_variable_allow_renaming('tau')
            tau.set_rhs(myokit.Divide(myokit.Number(1), myokit.Plus(a, b)))
            inf = x.add_variable_allow_renaming('inf')
            inf.set_rhs(myokit.Multiply(a, myokit.Name(tau)))

            # Update RHS expression for state
            x.set_rhs(myokit.Divide(
                myokit.Minus(myokit.Name(inf), myokit.Name(x)),
                myokit.Name(tau)
            ))

    return model


def get_alpha_and_beta(x, v=None):
    """
    Tests if the given ``x`` is a state variable with an expression of the form
    ``(1 - x) * alpha - x * beta``, and returns the variables for ``alpha`` and
    ``beta`` if so.

    Here, ``alpha(v)`` and ``beta(v)`` represent the forward and backward
    reaction rates for ``x``. Both may depend on ``v``, but not on any (other)
    state variable.

    Note that this method performs a shallow check of the equation's shape,
    and does not perform any simplification or rewriting to see if the
    expression can be made to fit the required form.

    Arguments:

    ``x``
        The :class:`myokit.Variable` to check.
    ``v``
        An optional :class:`myokit.Variable` representing the membrane
        potential. If not given, the label ``membrane_potential`` will be used
        to determine ``v``. If ``v=None`` and no membrane potential can be
        found an error will be raised. Membrane potential is typically
        specified as a state, but this is not a requirement.

    Returns a tuple ``(alpha, beta)`` if successful, or ``None`` if not. Both
    ``alpha`` and ``beta`` are :class:`myokit.Variable` objects.

    """
    #TODO: This method _could_ perhaps be made to allow arbitrary inline
    # expressions for alpha and beta, so that it would recognise that e.g.
    # (1 - x) * 0.5 - x * log(3) is suitable. Guessing this could work if you
    # collected x terms to get e1 - e2 * x, and then use e1 = alpha, and
    # e2 = alpha + beta = e1 + beta; beta = e2 - e1.
    # We could then add new variables for alpha and beta, and return those.

    # Check that ``x`` is a state
    if not x.is_state():
        return None

    # Check shape of RHS is (1 - x) * alpha - beta * x
    rhs = x.rhs()
    if not isinstance(rhs, myokit.Minus):
        return None
    a, b = rhs
    if not (isinstance(a, myokit.Multiply) and isinstance(b, myokit.Multiply)):
        return None

    # Test alpha
    xname = myokit.Name(x)
    xconv = myokit.Minus(myokit.Number(1), xname)
    if a[0] == xconv:
        alpha = a[1]
    elif a[1] == xconv:
        alpha = a[0]
    else:
        return None

    # ...and beta
    if b[0] == xname:
        beta = b[1]
    elif b[1] == xname:
        beta = b[0]
    else:
        return None

    # Check that alpha and beta are variable references
    if not (isinstance(alpha, myokit.Name) and isinstance(beta, myokit.Name)):
        return None

    # Check that alpha and beta are not states
    alpha = alpha.var()
    beta = beta.var()
    if alpha.is_state() or beta.is_state():
        return None

    # Check that, with the possible exception of v, alpha and beta don't depend
    # on any states

    # Get model from state varible
    model = x.model()

    # Check membrane potential variable is known.
    if v is None:
        v = model.label('membrane_potential')
    else:
        # Ensure v is a variable, and from the cloned model
        if isinstance(v, myokit.Variable):
            v = v.qname()
        v = model.get(v)
    if v is None:
        raise ValueError(
            'Membrane potential must be given as `v` or by setting the'
            ' label `membrane_potential` in the model.')

    # Check alpha and beta
    for term in (alpha, beta):

        # Get rhs, with any references to other variables inlined
        rhs = term.rhs().clone(expand=True, retain=(v, ))

        # Check that alpha and beta are not functions of any states (with the
        # possible exception of v).
        for state in model.states():
            if state != v and rhs.depends_on(myokit.Name(state)):
                return None

    # Looks good!
    return (alpha, beta)


def get_inf_and_tau(x, v=None):
    """
    Tests if the given ``x`` is a state variable with an expression of the form
    ``(x_inf - x) / tau_x``, and returns the variables for ``x_inf`` and
    ``x_tau`` if so.

    Here, ``x_inf`` and ``tau_x`` represent the steady-state and time
    constant of ``x``. Both may depend on ``v``, but not on any (other) state
    variable.

    Note that this method performs a shallow check of the equation's shape,
    and does not perform any simplification or rewriting to see if the
    expression can be made to fit the required form.

    Arguments:

    ``x``
        The :class:`myokit.Variable` to check.
    ``v``
        An optional :class:`myokit.Variable` representing the membrane
        potential. If not given, the label ``membrane_potential`` will be used
        to determine ``v``. If ``v=None`` and no membrane potential can be
        found an error will be raised. Membrane potential is typically
        specified as a state, but this is not a requirement.

    Returns:

    A tuple ``(x_inf, tau_x)`` if successful, or ``None`` if not. The returned
    ``x_inf`` and ``tau_x`` are :class:`myokit.Variable` objects.

    """
    #TODO: This method _could_ perhaps be made to accept any expression that
    # can be written in the form ``e1 - x / e2``, so that e2 = tau, and
    # e1*e2 = inf.
    # We could then add new variables for tau and inf, and return those.

    # Check that ``x`` is a state
    if not x.is_state():
        return None

    # Check shape of RHS is (x_inf - x) / tau_x
    rhs = x.rhs()
    if not isinstance(rhs, myokit.Divide):
        return None
    numer, xtau = rhs
    if not isinstance(xtau, myokit.Name):
        return None
    if not isinstance(numer, myokit.Minus):
        return None
    xinf, xname = numer
    if not (isinstance(xinf, myokit.Name) and isinstance(xname, myokit.Name)):
        return None
    if not xname == myokit.Name(x):
        return None

    # Check that inf and tau are not states
    xinf = xinf.var()
    xtau = xtau.var()
    if xinf.is_state() or xtau.is_state():
        return None

    # Check that, with the possible exception of v, xinf and xtau do not depend
    # on any states
    # Get model from state varible
    model = x.model()

    # Check membrane potential variable is known.
    if v is None:
        v = model.label('membrane_potential')
    else:
        # Ensure v is a variable, and from the cloned model
        if isinstance(v, myokit.Variable):
            v = v.qname()
        v = model.get(v)
    if v is None:
        raise ValueError(
            'Membrane potential must be given as `v` or by setting the'
            ' label `membrane_potential` in the model.')

    # Check xinf and xtau
    for term in (xinf, xtau):

        # Get rhs, with any references to other variables inlined
        rhs = term.rhs().clone(expand=True, retain=(v,))

        # Check if x_inf and tau_x are not functions of any states (with
        # the possible exception of v).
        for state in model.states():
            if state != v and rhs.depends_on(myokit.Name(state)):
                return None

    # Looks good!
    return xinf, xtau


def get_rl_expression(x, dt, v=None):
    """
    For states ``x`` with RHS expressions written in the "inf-tau form" (see
    :meth:`has_inf_tau_form`) this returns a Rush-Larsen (RL) update expression
    of the form ``x_inf + (x - x_inf) * exp(-(dt / tau_x))``.

    If the state does not have the "inf-tau form", ``None`` is returned.

    Arguments:

    ``x``
        A state variable for which to return the RL update expression.
    ``dt``
        A :class:`myokit.Name` expression to use for the time step in the RL
        expression.
    ``v``
        An optional variable representing the membrane potential.

    Returns a :class:`myokit.Expression` if succesful, or ``None`` if not.

    Example:

        # Load a Myokit model
        model = myokit.load_model('example')
        v = model.get('membrane.V')

        # Get a copy where all HH-state variables are written in inf-tau form
        model = myokit.lib.hh.convert_hh_states_to_inf_tau_form(model, v)

        # Create an expression for the time step
        dt = myokit.Name('dt')

        # Show an RL-update for the variable ina.h
        h = model.get('ina.h')
        e = myokit.lib.hh.get_rl_expression(h, dt, v)
        print('h[t + dt] = ' + str(e))

    See:

    [1] Rush, Larsen (1978) A Practical Algorithm for Solving Dynamic Membrane
    Equations. IEEE Transactions on Biomedical Engineering,
    https://doi.org/10.1109/TBME.1978.326270

    [2] Marsh, Ziaratgahi, Spiteri (2012) The secrets to the success of the
    Rush-Larsen method and its generalization. IEEE Transactions on Biomedical
    Engineering, https://doi.org/10.1109/TBME.2012.2205575

    """
    # Test dt is an expression
    if not isinstance(dt, myokit.Expression):
        raise ValueError('Argument `dt` must be a myokit.Expression.')

    # Get x_inf and tau_x, if possible
    res = get_inf_and_tau(x, v)
    if res is None:
        return None

    # Create expression for RL update
    # x_inf + (x - x_inf) * exp(-(dt / tau_x))
    x = myokit.Name(x)
    x_inf = myokit.Name(res[0])
    tau_x = myokit.Name(res[1])
    return myokit.Plus(
        x_inf,
        myokit.Multiply(
            myokit.Minus(x, x_inf),
            myokit.Exp(myokit.PrefixMinus(myokit.Divide(dt, tau_x)))
        )
    )


def has_alpha_beta_form(x, v=None):
    """
    Tests if the given ``x`` is a state variable with an expression of the form
    ``(1 - x) * alpha - x * beta``, where ``alpha`` and ``beta`` represent the
    forward and backward reaction rates for ``x``.

    If the optional argument ``v`` is given, the method will (1) check that
    both ``alpha`` and ``beta`` depend on ``v``, and (2) check that they don't
    depend on any state variables (not counting ``v``).

    Note that this method performs a shallow check of the equation's shape,
    and does not perform any simplification or rewriting to see if the
    expression can be made to fit the required form.

    Arguments:

    ``x``
        The :class:`myokit.Variable` to check.
    ``v``
        An optional :class:`myokit.Variable` representing the membrane
        potential.

    Returns ``True`` if the alpha-beta form is found, ``False`` if not.
    """
    try:
        alpha, beta = get_alpha_and_beta(x, v)
        return True
    except TypeError:
        return False


def has_inf_tau_form(x, v=None):
    """
    Tests if the given ``x`` is a state variable with an expression of the form
    ``(x_inf - x) / tau_x``, where ``x_inf`` and ``tau_x`` represent the
    steady-state and time constant of ``x``.

    If the optional argument ``v`` is given, the method will (1) check that
    both ``x_inf`` and ``tau_x`` depend on ``v``, and (2) check that they don't
    depend on any state variables (not counting ``v``).

    Note that this method performs a shallow check of the equation's shape,
    and does not perform any simplification or rewriting to see if the
    expression can be made to fit the required form.

    Arguments:

    ``x``
        The :class:`myokit.Variable` to check.
    ``v``
        An optional :class:`myokit.Variable` representing the membrane
        potential.

    Returns ``True`` if the inf-tau form is found, ``False`` if not.
    """
    try:
        x_inf, tau_x = get_inf_and_tau(x, v)
        return True
    except TypeError:
        return False


class HHModelError(myokit.MyokitError):
    """
    Raised for issues with constructing or using a :class:`LinearModel`.
    """




#
# CellML 2.0 support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._api import (     # noqa
    AnnotatableElement,
    CellMLError,
    clean_identifier,
    create_unit_name,
    Component,
    Model,
    Units,
    Variable,
    is_basic_real_number_string,
    is_identifier,
    is_integer_string,
    is_real_number_string,
)

from ._parser import (  # noqa
    parse_file,
    parse_string,
    CellMLParser,
    CellMLParsingError,
)

from ._writer import (  # noqa
    write_file,
    write_string,
    CellMLWriter,
)




#!/usr/bin/env python3
#
# Tests the Simulation1d class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import DIR_DATA, CancellingReporter

# Show simulation output
debug = False


class Simulation1dTest(unittest.TestCase):
    """
    Test the non-parallel 1d simulation.
    """

    def test_basic(self):
        # Test basic usage.

        # Load model
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Run a simulation
        ncells = 5
        s = myokit.Simulation1d(m, p, ncells=ncells)
        s.set_step_size(0.05)

        x0 = m.initial_values(True)
        self.assertEqual(s.time(), 0)
        self.assertEqual(s.state(0), x0)
        self.assertEqual(s.default_state(0), x0)
        d = s.run(5, log_interval=1)
        self.assertEqual(s.time(), 5)
        self.assertNotEqual(s.state(0), x0)
        self.assertEqual(s.default_state(0), x0)

        # Test full state getting and reset
        self.assertEqual(s.default_state(), x0 * ncells)
        self.assertNotEqual(s.state(), x0 * ncells)
        s.reset()
        self.assertEqual(s.state(), s.default_state())

        # Test pre updates the default state.
        s.pre(1)
        self.assertNotEqual(s.default_state(0), x0)

        # Simulation time can't be negative
        self.assertRaises(ValueError, s.run, -1)

        # Number of cells must be >0
        self.assertRaisesRegex(
            ValueError, 'number of cells', myokit.Simulation1d, m, p, 0)

        # Model must have a membrane potential
        v = m.get('membrane.V')
        v.set_label(None)
        self.assertRaisesRegex(
            ValueError, 'membrane_potential', myokit.Simulation1d, m, p, 5)
        v.set_label('membrane_potential')

        # Model must have a diffusion current
        d = m.binding('diffusion_current')
        d.set_binding(None)
        self.assertRaisesRegex(
            ValueError, 'diffusion_current', myokit.Simulation1d, m, p, 5)

        # Test setting conductance
        s.set_conductance(10)
        self.assertEqual(s.conductance(), 10)
        self.assertRaises(ValueError, s.set_conductance, -1)

        # Test setting paced cells
        s.set_paced_cells(1)
        self.assertEqual(s.paced_cells(), 1)
        self.assertRaises(ValueError, s.set_paced_cells, -1)

        # Test setting step size
        s.set_step_size(0.01)
        self.assertRaises(ValueError, s.set_step_size, 0)

        # Test setting time
        self.assertNotEqual(s.time(), 100)
        s.set_time(100)
        self.assertEqual(s.time(), 100)

    def test_initial_value_expressions(self):
        # Test if initial value expressions are converted to floats
        m = myokit.parse_model('''
            [[model]]
            c.x = 1 + sqrt(3)
            c.y = 1 / c.p
            c.z = 3

            [c]
            t = 0 bind time
            dot(x) = 1 label membrane_potential
            dot(y) = 2
            dot(z) = 3
            p = log(3)
            q = 0 bind diffusion_current
        ''')
        s = myokit.Simulation1d(m, ncells=2)
        x = s.state()
        self.assertIsInstance(x[0], float)
        self.assertIsInstance(x[1], float)
        self.assertIsInstance(x[2], float)
        self.assertEqual(x[:3], x[3:])
        self.assertEqual(x, m.initial_values(True) * 2)
        self.assertEqual(x, s.default_state())

    def test_negative_time(self):
        # Test starting at a negative time

        m = myokit.load_model(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        p = myokit.pacing.blocktrain(duration=1, level=1, period=2)
        n = 4
        s = myokit.Simulation1d(m, p, ncells=n)

        s.set_time(-10)
        self.assertEqual(s.time(), -10)
        d = s.run(5, log=['engine.pace']).npview()
        self.assertEqual(s.time(), -5)
        self.assertTrue(np.all(d['engine.pace'] == 0))
        d = s.run(6, log=['engine.time', 'engine.pace']).npview()
        #print(d['engine.time'])
        #print(d['engine.pace'])
        self.assertEqual(s.time(), 1)
        self.assertTrue(np.all(d['engine.pace'][:-1] == 0))
        self.assertEqual(d['engine.pace'][-1], 1)

    def test_no_protocol(self):
        # Test running without a protocol
        m = myokit.load_model(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        x = 0.123
        m.get('engine.pace').set_rhs(x)
        n = 4

        # Create without a protocol
        s = myokit.Simulation1d(m, ncells=n)
        d = s.run(10, log=['engine.pace']).npview()
        self.assertIn('engine.pace', d.keys())
        self.assertTrue(np.all(d['engine.pace'] == 0))

        # Set, then unset
        y = 0.101
        p = myokit.pacing.blocktrain(level=y, duration=100, period=100)
        s.set_protocol(p)
        d = s.run(5, log=['engine.pace']).npview()
        self.assertTrue(np.all(d['engine.pace'] == y))
        self.assertGreater(len(d['engine.pace']), 1)
        s.set_protocol(None)
        d = s.run(5, log=['engine.pace']).npview()
        self.assertTrue(np.all(d['engine.pace'] == 0))

    def test_set_state(self):
        # Test :meth:`Simulation1d.set_state()`.
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        n = 4

        s = myokit.Simulation1d(m, p, n)
        x0 = m.initial_values(True)
        self.assertEqual(s.state(), x0 * n)

        # Test setting a full state
        sx = [0] * 8 * n
        self.assertNotEqual(sx, s.state())
        s.set_state(sx)
        self.assertEqual(sx, s.state())

        # Test setting a single, global state
        sx = [0] * 8
        s.set_state(sx)
        self.assertEqual(s.state(), sx * n)
        s.set_state(x0)
        self.assertEqual(s.state(), x0 * n)

        # Test setting a single state
        j = 1
        s.set_state(sx, j)
        for i in range(n):
            if i == j:
                self.assertEqual(s.state(i), sx)
            else:
                self.assertEqual(s.state(i), x0)

        # Invalid cell index
        s.set_state(sx, 0)
        self.assertRaises(ValueError, s.set_state, sx, -1)
        self.assertRaises(ValueError, s.set_state, sx, n)
        self.assertRaises(ValueError, s.state, -1)
        self.assertRaises(ValueError, s.state, n)

    def test_set_default_state(self):
        # Test :meth:`Simulation1d.set_default_state()`.
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        n = 4

        s = myokit.Simulation1d(m, p, n)
        self.assertEqual(s.state(), m.initial_values(True) * n)

        # Test setting a full state
        sx = [0] * 8 * n
        self.assertNotEqual(sx, s.default_state())
        s.set_default_state(sx)
        self.assertEqual(sx, s.default_state())

        # Test setting a single, global state
        sx = [0] * 8
        s.set_default_state(sx)
        self.assertEqual(s.default_state(), sx * n)
        sy = m.initial_values(True)
        s.set_default_state(sy)
        self.assertEqual(s.default_state(), sy * n)

        # Test setting a single state
        j = 1
        s.set_default_state(sx, j)
        for i in range(n):
            if i == j:
                self.assertEqual(s.default_state(i), sx)
            else:
                self.assertEqual(s.default_state(i), sy)

        # Invalid cell index
        s.set_default_state(sx, 0)
        self.assertRaises(ValueError, s.set_default_state, sx, -1)
        self.assertRaises(ValueError, s.set_default_state, sx, n)
        self.assertRaises(ValueError, s.default_state, -1)
        self.assertRaises(ValueError, s.default_state, n)

    def test_with_progress_reporter(self):
        # Test running with a progress reporter.
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Test using a progress reporter
        s = myokit.Simulation1d(m, p, ncells=5)
        s.set_step_size(0.05)
        with myokit.tools.capture() as c:
            s.run(110, progress=myokit.ProgressPrinter())
        c = c.text().splitlines()
        self.assertTrue(len(c) > 0)

        # Not a progress reporter
        self.assertRaisesRegex(
            ValueError, 'ProgressReporter', s.run, 5, progress=12)

        # Cancel from reporter
        self.assertRaises(
            myokit.SimulationCancelledError, s.run, 1,
            progress=CancellingReporter(0))

    def test_against_cvode(self):
        # Compare the Simulation1d output with CVODE output

        # Load model and event with appropriate level/duration
        m, p, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))
        e = p.head()

        #
        # Make protocol to compare t=0 event implementations
        #
        dt = 0.1    # Note: this is too big to be very accurate
        tmax = 1300
        p = myokit.Protocol()
        p.schedule(level=e.level(), duration=e.duration(), period=600, start=0)
        logvars = ['engine.time', 'membrane.V', 'engine.pace', 'ica.ICa']

        s1 = myokit.Simulation1d(m, p, ncells=1, rl=True)
        s1.set_step_size(dt)
        d1 = s1.run(tmax, logvars, log_interval=dt).npview()

        s2 = myokit.Simulation(m, p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(tmax, logvars, log_interval=dt).npview()

        # Check implementation of logging point selection
        e0 = np.max(np.abs(d1.time() - d2.time()))

        # Check implementation of pacing
        r1 = d1['engine.pace'] - d2['engine.pace']
        e1 = np.sum(r1**2)

        # Check membrane potential (will have some error!)
        # Using MRMS from Marsh, Ziaratgahi, Spiteri 2012
        r2 = d1['membrane.V', 0] - d2['membrane.V']
        r2 /= (1 + np.abs(d2['membrane.V']))
        e2 = np.sqrt(np.sum(r2**2) / len(r2))

        # Check logging of intermediary variables
        r3 = d1['ica.ICa', 0] - d2['ica.ICa']
        r3 /= (1 + np.abs(d2['ica.ICa']))
        e3 = np.sqrt(np.sum(r3**2) / len(r3))

        if debug:
            import matplotlib.pyplot as plt
            print('Event at t=0')

            plt.figure()
            plt.suptitle('Time points')
            plt.plot(d1.time(), label='Euler')
            plt.plot(d1.time(), label='CVODE')
            plt.legend()
            print(d1.time()[:7])
            print(d2.time()[:7])
            print(d1.time()[-7:])
            print(d2.time()[-7:])
            print(e0)

            plt.figure()
            plt.suptitle('Pacing signals')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['engine.pace'], label='Euler')
            plt.plot(d2.time(), d2['engine.pace'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r1)
            print(e1)

            plt.figure()
            plt.suptitle('Membrane potential')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['membrane.V', 0], label='Euler')
            plt.plot(d2.time(), d2['membrane.V'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e2)

            plt.figure()
            plt.suptitle('Calcium current')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['ica.ICa', 0], label='Euler')
            plt.plot(d2.time(), d2['ica.ICa'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e3)

            plt.show()

        self.assertLess(e0, 1e-14)
        self.assertLess(e1, 1e-14)
        self.assertLess(e2, 0.1)   # Note: The step size is really too big here
        self.assertLess(e3, 0.1)

        #
        # Make protocol to compare with event at step-size multiple
        #
        dt = 0.1   # Note: this is too big to be very accurate
        tmax = 1300
        p = myokit.Protocol()
        p.schedule(
            level=e.level(),
            duration=e.duration(),
            period=600,
            start=10)

        logvars = ['engine.time', 'membrane.V', 'engine.pace', 'ica.ICa']

        s1.reset()
        s1.set_protocol(p)
        s1.set_step_size(dt)
        d1 = s1.run(tmax, logvars, log_interval=dt).npview()

        s2.reset()
        s2.set_protocol(p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(tmax, logvars, log_interval=dt).npview()

        # Check implementation of logging point selection
        e0 = np.max(np.abs(d1.time() - d2.time()))

        # Check implementation of pacing
        r1 = d1['engine.pace'] - d2['engine.pace']
        e1 = np.sum(r1**2)

        # Check membrane potential (will have some error!)
        # Using MRMS from Marsh, Ziaratgahi, Spiteri 2012
        r2 = d1['membrane.V', 0] - d2['membrane.V']
        r2 /= (1 + np.abs(d2['membrane.V']))
        e2 = np.sqrt(np.sum(r2**2) / len(r2))

        # Check logging of intermediary variables
        r3 = d1['ica.ICa', 0] - d2['ica.ICa']
        r3 /= (1 + np.abs(d2['ica.ICa']))
        e3 = np.sqrt(np.sum(r3**2) / len(r3))

        if debug:
            import matplotlib.pyplot as plt
            print('Event at t=0')

            plt.figure()
            plt.suptitle('Time points')
            plt.plot(d1.time(), label='Euler')
            plt.plot(d1.time(), label='CVODE')
            plt.legend()
            print(d1.time()[:7])
            print(d2.time()[:7])
            print(d1.time()[-7:])
            print(d2.time()[-7:])
            print(e0)

            plt.figure()
            plt.suptitle('Pacing signals')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['engine.pace'], label='Euler')
            plt.plot(d2.time(), d2['engine.pace'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r1)
            print(e1)

            plt.figure()
            plt.suptitle('Membrane potential')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['membrane.V', 0], label='Euler')
            plt.plot(d2.time(), d2['membrane.V'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e2)

            plt.figure()
            plt.suptitle('Calcium current')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['ica.ICa', 0], label='Euler')
            plt.plot(d2.time(), d2['ica.ICa'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e3)

            plt.show()

        self.assertLess(e0, 1e-14)
        self.assertLess(e1, 1e-14)
        self.assertLess(e2, 0.1)   # Note: The step size is really too big here
        self.assertLess(e3, 0.1)

        #
        # Make protocol to compare with event NOT at step-size multiple
        #
        dt = 0.1   # Note: this is too big to be very accurate
        tmax = 1300
        p = myokit.Protocol()
        p.schedule(
            level=e.level(),
            duration=e.duration(),
            period=600,
            start=1.05)

        logvars = ['engine.time', 'membrane.V', 'engine.pace', 'ica.ICa']

        s1.reset()
        s1.set_protocol(p)
        s1.set_step_size(dt)
        d1 = s1.run(tmax, logvars, log_interval=dt).npview()

        s2.reset()
        s2.set_protocol(p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(tmax, logvars, log_interval=dt).npview()

        # Check implementation of logging point selection
        e0 = np.max(np.abs(d1.time() - d2.time()))

        # Check implementation of pacing
        r1 = d1['engine.pace'] - d2['engine.pace']
        e1 = np.sum(r1**2)

        # Check membrane potential (will have some error!)
        # Using MRMS from Marsh, Ziaratgahi, Spiteri 2012
        r2 = d1['membrane.V', 0] - d2['membrane.V']
        r2 /= (1 + np.abs(d2['membrane.V']))
        e2 = np.sqrt(np.sum(r2**2) / len(r2))

        # Check logging of intermediary variables
        r3 = d1['ica.ICa', 0] - d2['ica.ICa']
        r3 /= (1 + np.abs(d2['ica.ICa']))
        e3 = np.sqrt(np.sum(r3**2) / len(r3))

        if debug:
            import matplotlib.pyplot as plt
            print('Event at t=0')

            plt.figure()
            plt.suptitle('Time points')
            plt.plot(d1.time(), label='Euler')
            plt.plot(d1.time(), label='CVODE')
            plt.legend()
            print(d1.time()[:7])
            print(d2.time()[:7])
            print(d1.time()[-7:])
            print(d2.time()[-7:])
            print(e0)

            plt.figure()
            plt.suptitle('Pacing signals')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['engine.pace'], label='Euler')
            plt.plot(d2.time(), d2['engine.pace'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r1)
            print(e1)

            plt.figure()
            plt.suptitle('Membrane potential')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['membrane.V', 0], label='Euler')
            plt.plot(d2.time(), d2['membrane.V'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e2)

            plt.figure()
            plt.suptitle('Calcium current')
            plt.subplot(2, 1, 1)
            plt.plot(d1.time(), d1['ica.ICa', 0], label='Euler')
            plt.plot(d2.time(), d2['ica.ICa'], label='CVODE')
            plt.legend()
            plt.subplot(2, 1, 2)
            plt.plot(d1.time(), r2)
            print(e3)

            plt.show()

        self.assertLess(e0, 1e-14)
        self.assertLess(e1, 1e-14)
        self.assertLess(e2, 0.2)   # Note: The step size is really too big here
        self.assertLess(e3, 0.2)

        #
        # Test again, with event at step multiple, but now without Rush-Larsen
        #
        dt = 0.01   # Note: this is too big to be very accurate
        tmax = 200
        p = myokit.Protocol()
        p.schedule(
            level=e.level(),
            duration=e.duration(),
            period=600,
            start=10)

        logvars = ['engine.time', 'membrane.V', 'engine.pace', 'ica.ICa']
        s1 = myokit.Simulation1d(m, p, ncells=1, rl=False)
        s1.set_step_size(dt)
        d1 = s1.run(tmax, logvars, log_interval=dt).npview()
        s2.reset()
        s2.set_protocol(p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(tmax, logvars, log_interval=dt).npview()

        # Check membrane potential (will have some error!)
        # Using MRMS from Marsh, Ziaratgahi, Spiteri 2012
        r2 = d1['membrane.V', 0] - d2['membrane.V']
        r2 /= (1 + np.abs(d2['membrane.V']))
        e2 = np.sqrt(np.sum(r2**2) / len(r2))
        self.assertLess(e2, 0.05)  # Note: The step size is really too big here

        # Check logging of intermediary variables
        r3 = d1['ica.ICa', 0] - d2['ica.ICa']
        r3 /= (1 + np.abs(d2['ica.ICa']))
        e3 = np.sqrt(np.sum(r3**2) / len(r3))
        self.assertLess(e3, 0.05)


if __name__ == '__main__':
    import sys
    if '-v' in sys.argv:
        print('Running in debug/verbose mode')
        debug = True
    else:
        print('Add -v for more debug output')
    unittest.main()



#
# Myokit auxillary functions: This module can be used to gather any
# functions that are important enough to warrant inclusion in the main
# myokit module but don't belong to any specific hidden module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import array
import io
import os
import sys

import myokit


def _examplify(filename):
    """
    If ``filename`` is equal to "example" and there isn't a file with that
    name, this function returns the file specified by myokit.EXAMPLE. In all
    other cases, the original filename is returned.
    """
    if filename == 'example' and not os.path.exists(filename):
        return myokit.EXAMPLE
    else:
        return os.path.expanduser(filename)


def load(filename):
    """
    Reads an ``mmt`` file and returns a tuple ``(model, protocol, embedded
    script)``.

    If the file specified by ``filename`` doesn't contain one of these parts
    the corresponding entry in the tuple will be ``None``.
    """
    f = open(_examplify(filename), 'r')
    try:
        return myokit.parse(f)
    finally:
        f.close()


def load_model(filename):
    """
    Loads the model section from an ``mmt`` file.

    Raises a :class:`SectionNotFoundError` if no model section is found.
    """
    filename = _examplify(filename)
    with open(filename, 'r') as f:
        section = myokit.split(f)[0]
        if not section.strip():
            raise myokit.SectionNotFoundError('Model section not found.')
        return myokit.parse(section.splitlines())[0]


def load_protocol(filename):
    """
    Loads the protocol section from an ``mmt`` file.

    Raises a :class:`SectionNotFoundError` if no protocol section is found.
    """
    filename = _examplify(filename)
    with open(filename, 'r') as f:
        section = myokit.split(f)[1]
        if not section.strip():
            raise myokit.SectionNotFoundError('Protocol section not found.')
        return myokit.parse(section.splitlines())[1]


def load_script(filename):
    """
    Loads the script section from an ``mmt`` file.

    Raises a :class:`SectionNotFoundError` if no script section is found.
    """
    filename = _examplify(filename)
    with open(filename, 'r') as f:
        section = myokit.split(f)[2]
        if not section.strip():
            raise myokit.SectionNotFoundError('Script section not found.')
        return myokit.parse(section.splitlines())[2]


def load_state(filename, model=None):
    """
    Loads a model state from a file in one of the formats specified by
    :func:`myokit.parse_state()`.

    If a :class:`Model` is provided the state will be run through
    :meth:`Model.map_to_state()` and returned as a list of floating point
    numbers.
    """
    filename = os.path.expanduser(filename)
    with open(filename, 'r') as f:
        s = myokit.parse_state(f)
        if model:
            s = model.map_to_state(s)
        return s


def load_state_bin(filename):
    """
    Loads a model state from a file in the binary format used by Myokit.

    See :meth:`save_state_bin` for details.
    """
    filename = os.path.expanduser(filename)

    # Load compression modules
    import zipfile
    try:
        import zlib
        del zlib
    except ImportError:
        raise Exception(
            'This method requires the `zlib` module to be installed.')

    # Open file
    with zipfile.ZipFile(filename, 'r') as f:
        info = f.infolist()

        if len(info) != 1:  # pragma: no cover
            raise Exception('Invalid state file format [10].')

        # Split into parts, get data type and array size
        info = info[0]
        parts = info.filename.split('_')

        if len(parts) != 3:     # pragma: no cover
            raise Exception('Invalid state file format [20].')

        if parts[0] != 'state':     # pragma: no cover
            raise Exception('Invalid state file format [30].')

        code = parts[1]
        if code not in ['d', 'f']:  # pragma: no cover
            raise Exception('Invalid state file format [40].')

        size = int(parts[2])
        if size < 0:    # pragma: no cover
            raise Exception('Invalid state file format [50].')

        # Create array, read bytes into array
        ar = array.array(code)
        ar.frombytes(f.read(info))

        # Always store as little endian
        if sys.byteorder == 'big':  # pragma: no cover
            ar.byteswap()

    return list(ar)


def save(filename=None, model=None, protocol=None, script=None):
    """
    Saves a model, protocol, and embedded script to an ``mmt`` file.

    The ``model`` argument can be given as plain text or a
    :class:`myokit.Model` object. Similarly, ``protocol`` can be either a
    :class:`myokit.Protocol` or its textual represenation.

    If no filename is given the ``mmt`` code is returned as a string.
    """
    if model is None and protocol is None and script is None:
        raise ValueError(
            'At least one of [model, protocol, script] must not be None.')

    if filename:
        filename = os.path.expanduser(filename)
        f = open(filename, 'w')
    else:
        f = io.StringIO()
    out = None
    try:
        if model is not None:
            if isinstance(model, myokit.Model):
                model = model.code()
            else:
                model = model.strip()
                if model != '' and model[:9] != '[[model]]':
                    f.write('[[model]]\n')
            model = model.strip()
            if model:
                f.write(model)
                f.write('\n\n')

        if protocol is not None:
            if isinstance(protocol, myokit.Protocol):
                protocol = protocol.code()
            else:
                protocol = protocol.strip()
                if protocol != '' and protocol[:12] != '[[protocol]]':
                    f.write('[[protocol]]\n')
            protocol = protocol.strip()
            if protocol:
                f.write(protocol)
                f.write('\n\n')

        if script is not None:
            script = script.strip()
            if script != '' and script[:10] != '[[script]]':
                f.write('[[script]]\n')
            if script:
                f.write(script)
                f.write('\n\n')
    finally:
        if filename:
            f.close()
        else:
            out = f.getvalue()
    return out


def save_model(filename, model):
    """
    Saves a model to a file
    """
    return save(filename, model)


def save_protocol(filename, protocol):
    """
    Saves a protocol to a file
    """
    return save(filename, protocol=protocol)


def save_script(filename, script):
    """
    Saves an embedded script to a file
    """
    return save(filename, script=script)


def save_state(filename, state, model=None):
    """
    Stores a model state to the path ``filename``.

    If no ``model`` is specified ``state`` should be given as a list of
    floating point numbers and will be stored by simply placing each number on
    a new line.

    If a :class:`Model <myokit.Model>` is provided the state can be in any
    format accepted by :meth:`Model.map_to_state() <myokit.Model.map_to_state>`
    and will be stored in the format returned by
    :meth:`Model.format_state() <myokit.Model.format_state>`.
    """
    # Check filename
    filename = os.path.expanduser(filename)

    # Format
    if model is not None:
        state = model.map_to_state(state)
        state = model.format_state(state)
    else:
        state = '\n'.join([myokit.float.str(s) for s in state])

    # Store
    with open(filename, 'w') as f:
        f.write(state)


def save_state_bin(filename, state, precision=myokit.DOUBLE_PRECISION):
    """
    Stores a model state (or any given list of floating point numbers) to the
    path ``filename``, using a binary format.

    The used format is a zip file, containing a single entry: ``state_x_y``,
    where ``x`` is the used data type (``d`` or ``f``) and ``y`` is the number
    of entries. All entries are stored little-endian.
    """
    # Check filename
    filename = os.path.expanduser(filename)

    # Load compression modules
    import zipfile
    try:
        import zlib
        del zlib
    except ImportError:
        raise Exception(
            'This method requires the `zlib` module to be installed.')

    # Data type
    code = 'd' if precision == myokit.DOUBLE_PRECISION else 'f'

    # Create array, ensure it's little-endian
    ar = array.array(code, state)
    if sys.byteorder == 'big':  # pragma: no cover
        ar.byteswap()

    # Store precision and data type in internal filename
    name = 'state_' + code + '_' + str(len(state))
    info = zipfile.ZipInfo(name)
    info.compress_type = zipfile.ZIP_DEFLATED

    # Write to compressed file
    ar = ar.tobytes()
    with zipfile.ZipFile(filename, 'w') as f:
        f.writestr(info, ar)




#
# C Compiler information class
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit

# Path to C Source for Compiler info module
SOURCE_FILE = 'compiler.c'


class Compiler(myokit.CModule):
    """
    Tests for distutils C-compilation support.
    """
    # Unique id for this object
    _index = 0

    # Cached back-end object if compiled, False if compilation failed
    _instance = None

    # Cached compilation error messages
    _message = None

    # Guess at compiler
    _compiler = None

    def __init__(self):
        super().__init__()
        # Create and cache back-end
        Compiler._index += 1

        # Define libraries
        libd = list()
        incd = list()
        incd.append(myokit.DIR_CFUNC)
        libs = []

        # Create back-end
        mname = 'myokit_compiler_info_' + str(Compiler._index)
        mname += '_' + str(myokit.pid_hash())
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)
        args = {'module_name': mname}
        try:
            Compiler._instance = self._compile(
                mname, fname, args, libs, libd, incd)
        except myokit.CompilationError as e:  # pragma: no cover
            Compiler._instance = False
            Compiler._message = str(e)

    @staticmethod
    def _get_instance():
        """
        Returns a cached back-end, creates and returns a new back-end or raises
        a :class:`NoCompilerError`.
        """
        # No instance? Create it
        if Compiler._instance is None:
            Compiler()

        # Instance creation failed, raise exception
        if Compiler._instance is False:  # pragma: no cover
            raise NoCompilerError(
                'Could not detect C compiler.\n' + Compiler._message)

        # Return instance
        return Compiler._instance

    @staticmethod
    def info(debug=False):
        """
        Returns a string with information about the compiler found on this
        system, or ``None`` if no compiler could be found.

        If ``debug`` is set to ``True``, compilation errors will be printed to
        stdout.
        """
        try:
            return Compiler._get_instance().compiler()
        except NoCompilerError as e:  # pragma: no cover
            if debug:
                print(e)
                if Compiler._message is None:
                    print('No further error message available.')
                else:
                    print(Compiler._message)
            return None


class NoCompilerError(myokit.MyokitError):
    """
    Raised when a compiler is required but not found.
    """

